import os
from decimal import Decimal

import pytest

import qcelemental


@pytest.mark.parametrize("inp", ["He100", '-1', -1, -1.0, 'cat', 200, 'Cro'])
def test_id_resolution_error_bad_element(inp):
    with pytest.raises(qcelemental.NotAnElementError):
        qcelemental.vdwradii.get(inp)


@pytest.mark.parametrize("inp", ['X', 'Fe', 100, 'Ru'])
def test_id_resolution_error(inp):
    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.vdwradii.get(inp)

    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.vdwradii.get(inp, return_tuple=True)

    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.vdwradii.get(inp, return_tuple=True, missing=3.0)

    assert qcelemental.vdwradii.get(inp, missing=4.0) == pytest.approx(4.0, 1.e-9)


a2b = 1. / qcelemental.constants.bohr2angstroms


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("KRYPTON", 2.02),
        ("kr", 2.02),
        ("kr84", 2.02),
        (36, 2.02),

        # Deuterium
        ("D", 1.10),
        ("h2", 1.10),

        # Germanium
        ("germanium", 2.11),
        ("Ge", 2.11),
    ])
def test_get(inp, expected):
    assert qcelemental.vdwradii.get(inp, units='angstrom') == pytest.approx(expected, 1.e-9)
    assert qcelemental.vdwradii.get(inp) == pytest.approx(a2b * expected, 1.e-9)


def test_get_tuple():
    ref = {'label': 'Mg', 'units': 'angstrom', 'data': Decimal('1.73')}
    dqca = qcelemental.vdwradii.get('magnesium', return_tuple=True).dict()

    for itm in ref:
        assert ref[itm] == dqca[itm]


def test_c_header():
    qcelemental.vdwradii.write_c_header("header.h")
    os.remove("header.h")


def test_representation():
    qcelemental.vdwradii.string_representation()


def test_str():
    print(str(qcelemental.vdwradii))
    assert "VanderWaalsRadii(" in str(qcelemental.vdwradii)


def test_vdwradmaker2018():
    with pytest.raises(KeyError) as e:
        qcelemental.VanderWaalsRadii("VDWRADMAKER2019")

    assert "only contexts {'MANTINA2009', } are currently supported" in str(e.value)



from copy import deepcopy
from typing import Any, Dict, Union

import numpy as np

from ..exceptions import ValidationError
from ..physical_constants import constants
from ..util import unnp
from .to_string import formula_generator


def to_schema(molrec: Dict[str, Any],
              dtype: Union[str, int],
              units: str = 'Bohr',
              *,
              np_out: bool = False,
              copy: bool = True) -> Dict[str, Any]:
    """Translate molparse internal Molecule spec into dictionary from other schemas.

    Parameters
    ----------
    molrec : dict
        Psi4 json Molecule spec.
    dtype : {'psi4', 1, 2}
        Molecule schema format.
        ``1`` is https://molssi-qc-schema.readthedocs.io/en/latest/auto_topology.html V1 + #44 + #53
        ``2`` is ``1`` with internal schema_name/version (https://github.com/MolSSI/QCSchema/pull/60)
    units : {'Bohr', 'Angstrom'}
        Units in which to write string. There is not an option to write in
        intrinsic/input units. Some `dtype` may not allow all units.
    np_out : bool, optional
        When `True`, fields originating from geom, elea, elez, elem, mass, real, elbl will be ndarray.
        Use `False` to get a json-able version.
    #return_type : {'json', 'yaml'} Serialization format string to return.

    Returns
    -------
    qcschema : dict
        Dictionary of the `dtype` repr of `molrec`.

    """
    qcschema: Dict = {}

    geom = np.array(molrec["geom"], copy=copy)

    if molrec["units"] == 'Bohr' and units == 'Bohr':
        pass
    elif molrec['units'] == 'Angstrom' and units == 'Bohr' and 'input_units_to_au' in molrec:
        geom = geom * molrec['input_units_to_au']
    else:
        geom = geom * constants.conversion_factor(molrec['units'], units)

    nat = geom.shape[0] // 3

    name = molrec.get('name', formula_generator(molrec['elem']))
    #    tagline = """auto-generated by qcdb from molecule {}""".format(name)

    if dtype == 'psi4':
        if units not in ['Angstrom', 'Bohr']:
            raise ValidationError("""Psi4 Schema {} allows only 'Bohr'/'Angstrom' coordinates, not {}.""".format(
                dtype, units))
        qcschema = deepcopy(molrec)
        qcschema['geom'] = geom
        qcschema['units'] = units
        qcschema['name'] = name

    elif dtype in [1, 2]:
        if units != 'Bohr':
            raise ValidationError("""QC_JSON_Schema {} allows only 'Bohr' coordinates, not {}.""".format(dtype, units))

        molecule: Dict = {}
        molecule["validated"] = True
        molecule['symbols'] = np.array(molrec['elem'], copy=copy)
        molecule['geometry'] = geom
        molecule['masses'] = np.array(molrec['mass'], copy=copy)
        molecule['atomic_numbers'] = np.array(molrec['elez'], copy=copy)
        molecule['mass_numbers'] = np.array(molrec['elea'], copy=copy)
        molecule['atom_labels'] = np.array(molrec['elbl'], copy=copy)
        molecule['name'] = name
        if 'comment' in molrec:
            molecule['comment'] = molrec['comment']
        molecule['molecular_charge'] = molrec['molecular_charge']
        molecule['molecular_multiplicity'] = molrec['molecular_multiplicity']
        molecule['real'] = np.array(molrec['real'], copy=copy)
        fidx = np.split(np.arange(nat), molrec['fragment_separators'])
        molecule['fragments'] = [fr.tolist() for fr in fidx]
        molecule['fragment_charges'] = np.array(molrec['fragment_charges']).tolist()
        molecule['fragment_multiplicities'] = np.array(molrec['fragment_multiplicities']).tolist()
        molecule['fix_com'] = molrec['fix_com']
        molecule['fix_orientation'] = molrec['fix_orientation']
        if 'fix_symmetry' in molrec:
            molecule['fix_symmetry'] = molrec['fix_symmetry']
        molecule['provenance'] = deepcopy(molrec['provenance'])
        if 'connectivity' in molrec:
            molecule['connectivity'] = deepcopy(molrec['connectivity'])

        if dtype == 1:
            qcschema = {'schema_name': 'qcschema_input', 'schema_version': 1, 'molecule': molecule}
        elif dtype == 2:
            qcschema = molecule
            qcschema.update({'schema_name': 'qcschema_molecule', 'schema_version': 2})

    else:
        raise ValidationError(
            "Schema dtype not understood, valid options are {{'psi4', 1, 2}}. Found {}.".format(dtype))

    if not np_out:
        qcschema = unnp(qcschema)

    return qcschema

    #if return_type == 'json':
    #    return json.dumps(qcschema)
    #elif return_type == 'yaml':
    #    import yaml
    #    return yaml.dump(qcschema)
    #else:
    #    raise ValidationError("""Return type ({}) not recognized.""".format(return_type))



import socket

import pytest

from qcelemental.util import which_import


def internet_connection():
    try:
        socket.create_connection(("www.google.com", 80))
        return True
    except OSError:
        return False


using_web = pytest.mark.skipif(internet_connection() is False, reason="Could not connect to the internet")

using_msgpack = pytest.mark.skipif(
    which_import('msgpack', return_bool=True) is False,
    reason='Not detecting module msgpack. Install package if necessary and add to envvar PYTHONPATH')

using_networkx = pytest.mark.skipif(
    which_import('networkx', return_bool=True) is False,
    reason='Not detecting module networkx. Install package if necessary and add to envvar PYTHONPATH')

using_scipy = pytest.mark.skipif(
    which_import('scipy', return_bool=True) is False,
    reason='Not detecting module scipy. Install package if necessary and add to envvar PYTHONPATH')

using_py3dmol = pytest.mark.skipif(
    which_import('py3Dmol', return_bool=True) is False,
    reason='Not detecting module py3Dmol. Install package if necessary and add to envvar PYTHONPATH')

serialize_extensions = ["json", "json-ext", pytest.param("msgpack-ext", marks=using_msgpack)]



"""
Datum Object Model
"""

from decimal import Decimal
from typing import Any, Dict, Optional

import numpy as np
from pydantic import BaseModel, validator


class Datum(BaseModel):
    """Facilitates the storage of quantum chemical results by labeling them with basic metadata.

    Attributes
    ----------
    label : str
        Official label for `data`, often qcvar. May contain spaces.
    units : str
        ASCII, LaTeX-like representation of units, without square brackets.
    data : float or Decimal or or :py:class:`numpy.ndarray`
        Value for `label`.
    comment : str, optional
        Additional notes.
    doi : str, optional
        Literature citation or definition DOI link.
    glossary : str, optional
        Extended description or definition.
    numeric : bool, optional
        Whether `data` is numeric. Pass `True` to disable validating `data` as float/Decimal/np.ndarray.

    """

    numeric: bool
    label: str
    units: str
    data: Any
    comment: str = ''
    doi: Optional[str] = None
    glossary: str = ''

    class Config:
        extra = "forbid"
        allow_mutation = False
        json_encoders = {
            np.ndarray: lambda v: v.flatten().tolist(),
            complex: lambda v: (v.real, v.imag),
        }

    def __init__(self, label, units, data, *, comment=None, doi=None, glossary=None, numeric=True):
        kwargs = {'label': label, 'units': units, 'data': data, 'numeric': numeric}
        if comment is not None:
            kwargs['comment'] = comment
        if doi is not None:
            kwargs['doi'] = doi
        if glossary is not None:
            kwargs['glossary'] = glossary

        super().__init__(**kwargs)

    @validator('data')
    def must_be_numerical(cls, v, values, **kwargs):
        try:
            1.0 * v
        except TypeError:
            try:
                Decimal('1.0') * v
            except TypeError:
                if values['numeric']:
                    raise ValueError(f'Datum data should be float, Decimal, or np.ndarray, not {type(v)}.')
            else:
                values['numeric'] = True
        else:
            values['numeric'] = True

        return v

    def __str__(self, label=''):
        width = 40
        text = ['-' * width, '{:^{width}}'.format('Datum ' + self.label, width=width)]
        if label:
            text.append('{:^{width}}'.format(label, width=width))
        text.append('-' * width)
        text.append('Data:     {}'.format(self.data))
        text.append('Units:    [{}]'.format(self.units))
        text.append('doi:      {}'.format(self.doi))
        text.append('Comment:  {}'.format(self.comment))
        text.append('Glossary: {}'.format(self.glossary))
        text.append('-' * width)
        return '\n'.join(text)

    def dict(self, *args, **kwargs):
        return super().dict(*args, **{**kwargs, **{"skip_defaults": True}})

    def to_units(self, units=None):
        from .physical_constants import constants

        to_unit = self.units if units is None else units
        factor = constants.conversion_factor(self.units, to_unit)

        if isinstance(self.data, Decimal):
            return factor * float(self.data)
        else:
            return factor * self.data


def print_variables(qcvars: Dict[str, 'Datum']) -> str:
    """Form a printable representation of qcvariables.

    Parameters
    ----------
    qcvars : dict of Datum
        Group of Datum objects to print.

    Returns
    -------
    str
        Printable string representation of label, data, and unit in Datum-s.

    """
    text = ['\n  Variable Map:', '  ----------------------------------------------------------------------------']

    if len(qcvars) == 0:
        text.append('  (none)')
        return '\n'.join(text)

    largest_key = max(len(k) for k in qcvars) + 2  # for quotation marks
    largest_characteristic = 8
    for k, v in qcvars.items():
        try:
            exp = int(str(v.data).split('E')[1])
        except IndexError:
            pass
        else:
            largest_characteristic = max(exp, largest_characteristic)

    for k, qca in sorted(qcvars.items()):
        #if k != qca.lbl:
        #    raise ValidationError('Huh? {} != {}'.format(k, qca.label))

        if isinstance(qca.data, np.ndarray):
            data = np.array_str(qca.data, max_line_width=120, precision=8, suppress_small=True)
            data = '\n'.join('        ' + ln for ln in data.splitlines())
            text.append("""  {:{keywidth}} => {:{width}} [{}]""".format('"' + k + '"',
                                                                        '',
                                                                        qca.units,
                                                                        keywidth=largest_key,
                                                                        width=largest_characteristic + 14))
            text.append(data)
        elif isinstance(qca.data, Decimal):
            text.append("""  {:{keywidth}} => {:{width}} [{}]""".format('"' + k + '"',
                                                                        qca.data,
                                                                        qca.units,
                                                                        keywidth=largest_key,
                                                                        width=largest_characteristic + 14))
        elif not qca.numeric:
            text.append("""  {:{keywidth}} => {:>{width}} [{}]""".format('"' + k + '"',
                                                                         str(qca.data),
                                                                         qca.units,
                                                                         keywidth=largest_key,
                                                                         width=largest_characteristic + 14))
        else:
            text.append("""  {:{keywidth}} => {:{width}.{prec}f} [{}]""".format('"' + k + '"',
                                                                                qca.data,
                                                                                qca.units,
                                                                                keywidth=largest_key,
                                                                                width=largest_characteristic + 14,
                                                                                prec=12))

    text.append('')
    return '\n'.join(text)



import pytest

import qcelemental

co_dominant = (59, 27, 'Co', 58.93319429, True, '')
co_dominant_mine = (59, 27, 'Co', 58.93319429, True, '_mine')
co_dominant_shortmass = (59, 27, 'Co', 58.933, True, '')
co60 = (60, 27, 'Co', 59.93381630, True, '')
co60_mine = (60, 27, 'Co', 59.93381630, True, '_mine')
co60ghost = (60, 27, 'Co', 59.93381630, False, '')
co_unspecified = (-1, 27, 'Co', 60.6, True, '')


@pytest.mark.parametrize("inp,expected", [
    ({'E': 'co'}, co_dominant),
    ({'Z': 27}, co_dominant),
    ({'A': 59, 'Z': 27}, co_dominant),
    ({'E': 'cO', 'mass': 58.93319429}, co_dominant),
    ({'A': 59, 'Z': 27, 'E': 'CO'}, co_dominant),
    ({'A': 59, 'E': 'cO', 'mass': 58.93319429}, co_dominant),
    ({'label': 'co', 'verbose': 2}, co_dominant),
    ({'label': '59co'}, co_dominant),
    ({'label': 'co@58.93319429'}, co_dominant),
    ({'A': 59, 'Z': 27, 'E': 'cO', 'mass': 58.93319429, 'label': 'co@58.93319429'}, co_dominant),
    ({'A': 59, 'Z': 27, 'E': 'cO', 'mass': 58.93319429, 'label': '27@58.93319429'}, co_dominant),
    ({'label': '27'}, co_dominant),
    ({'label': 'co_miNe'}, co_dominant_mine),
    ({'label': 'co_mIne@58.93319429'}, co_dominant_mine),
    ({'E': 'cO', 'mass': 58.933}, co_dominant_shortmass),
    ({'label': 'cO@58.933'}, co_dominant_shortmass),
    ({'E': 'Co', 'A': 60}, co60),
    ({'Z': 27, 'A': 60, 'real': True}, co60),
    ({'E': 'Co', 'A': 60}, co60),
    ({'Z': 27, 'mass': 59.93381630}, co60),
    ({'A': 60, 'Z': 27, 'mass': 59.93381630}, co60),
    ({'label': '60Co' }, co60),
    ({'label': '27', 'mass': 59.93381630}, co60),
    ({'label': 'Co', 'mass': 59.93381630}, co60),
    ({'A': 60, 'label': 'Co'}, co60),
    ({'mass': 60.6, 'Z': 27}, co_unspecified),
    ({'mass': 60.6, 'E': 'Co'}, co_unspecified),
    ({'mass': 60.6, 'label': '27'}, co_unspecified),
    ({'label': 'Co@60.6'}, co_unspecified),
    ({'E': 'Co', 'A': 60, 'real': False}, co60ghost),
    ({'A': 60, 'Z': 27, 'mass': 59.93381630, 'real': 0}, co60ghost),
    ({'label': '@60Co'}, co60ghost),
    ({'label': 'Gh(27)', 'mass': 59.93381630}, co60ghost),
    ({'label': '@Co', 'mass': 59.93381630}, co60ghost),
    ({'A': 60, 'label': 'Gh(Co)'}, co60ghost),
    ({'label': '60co_miNe'}, co60_mine),
    ({'label': '_miNe', 'A': 59, 'E': 'Co', 'speclabel': False}, co_dominant_mine),
])  # yapf: disable
def test_reconcile_nucleus(inp, expected):
    assert expected == qcelemental.molparse.reconcile_nucleus(**inp)


@pytest.mark.parametrize("inp", [
    {'E': 'cO', 'mass': 58.933, 'mtol': 1.e-4},
    {'label': '27@58.933', 'mtol': 1.e-4},
])  # yapf: disable
def test_reconcile_nucleus_assertionerror(inp):
    with pytest.raises(AssertionError):
        assert co_dominant_shortmass == qcelemental.molparse.reconcile_nucleus(**inp)


@pytest.mark.parametrize("inp", [
    {'A': 80, 'Z': 27 },
    {'Z': -27, 'mass': 200, 'nonphysical': True},
])  # yapf: disable
def test_reconcile_nucleus_notanelementerror(inp):
    with pytest.raises(qcelemental.NotAnElementError):
        qcelemental.molparse.reconcile_nucleus(**inp)


def test_reconcile_nucleus_41():
    qcelemental.molparse.reconcile_nucleus(Z=27, mass=200, nonphysical=True)


@pytest.mark.parametrize("inp", [
    {'mass': 60.6, 'Z': 27, 'A': 61},
    {'Z': 27, 'mass': 200 },
    {'Z': 27, 'mass': -200, 'nonphysical': True},
    {'Z': 1, 'label': 'he'},
    {'A': 4, 'label': '3he'},
    {'label': '@U', 'real': True},
    {'label': 'U', 'real': False},
    {'label': '_miNe', 'A': 59, 'E': 'Co', 'speclabel': True},
])  # yapf: disable
def test_reconcile_nucleus_validationerror(inp):
    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.reconcile_nucleus(**inp)



"""
Tests the imports and exports of the Molecule object.
"""

import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.models import Molecule
from qcelemental.testing import compare, compare_values

from .addons import serialize_extensions, using_msgpack, using_py3dmol

water_molecule = Molecule.from_data("""
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    """)

water_dimer_minima = Molecule.from_data("""
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    --
    O   1.350625   0.111469   0.000000
    H   1.680398  -0.373741  -0.758561
    H   1.680398  -0.373741   0.758561
    """,
                                        dtype="psi4",
                                        orient=True)


def test_molecule_data_constructor_numpy():
    water_psi = water_dimer_minima.copy()
    ele = np.array(water_psi.atomic_numbers).reshape(-1, 1)
    npwater = np.hstack((ele, water_psi.geometry * qcel.constants.conversion_factor("Bohr", "angstrom")))

    water_from_np = Molecule.from_data(npwater, name="water dimer", dtype="numpy", frags=[3])
    assert water_psi.compare(water_from_np)

    water_from_np = Molecule.from_data(npwater, name="water dimer", frags=[3])
    assert water_psi.compare(water_from_np)
    assert water_psi.get_molecular_formula() == "H4O2"


def test_molecule_data_constructor_dict():
    water_psi = water_dimer_minima.copy()

    # Check the JSON construct/deconstruct
    water_from_json = Molecule.from_data(water_psi.dict())
    assert water_psi.compare(water_from_json)

    water_from_json = Molecule.from_data(water_psi.json(), "json")
    assert water_psi.compare(water_from_json)
    assert water_psi.compare(Molecule.from_data(water_psi.to_string("psi4"), dtype="psi4"))

    assert water_psi.get_hash() == '3c4b98f515d64d1adc1648fe1fe1d6789e978d34'  # copied from schema_version=1
    assert water_psi.schema_version == 2
    assert water_psi.schema_name == 'qcschema_molecule'


def test_molecule_data_constructor_error():
    with pytest.raises(TypeError):
        Molecule.from_data([])

    with pytest.raises(KeyError):
        Molecule.from_data({}, dtype="bad")


def test_molecule_np_constructors():
    """
    Neon tetramer fun
    """
    ### Neon Tetramer
    neon_from_psi = Molecule.from_data("""
        Ne 0.000000 0.000000 0.000000
        --
        Ne 3.100000 0.000000 0.000000
        --
        Ne 0.000000 3.200000 0.000000
        --
        Ne 0.000000 0.000000 3.300000
        units bohr""",
                                       dtype="psi4")
    ele = np.array([10, 10, 10, 10]).reshape(-1, 1)
    npneon = np.hstack((ele, neon_from_psi.geometry))
    neon_from_np = Molecule.from_data(npneon, name="neon tetramer", dtype="numpy", frags=[1, 2, 3], units="bohr")

    assert neon_from_psi.compare(neon_from_np)

    # Check the JSON construct/deconstruct
    neon_from_json = Molecule.from_data(neon_from_psi.json(), dtype="json")
    assert neon_from_psi.compare(neon_from_json)
    assert neon_from_json.get_molecular_formula() == "Ne4"


def test_molecule_compare():
    water_molecule2 = water_molecule.copy()
    assert water_molecule2.compare(water_molecule)

    water_molecule3 = water_molecule.copy(update={"geometry": (water_molecule.geometry + np.array([0.1, 0, 0]))})
    assert water_molecule.compare(water_molecule3) is False


def test_water_minima_data():
    # Give it a name
    mol_dict = water_dimer_minima.dict()
    mol_dict['name'] = "water dimer"
    mol = Molecule(orient=True, **mol_dict)

    assert len(str(mol)) == 661
    assert len(mol.to_string("psi4")) == 479

    assert sum(x == y for x, y in zip(mol.symbols, ['O', 'H', 'H', 'O', 'H', 'H'])) == mol.geometry.shape[0]
    assert mol.name == "water dimer"
    assert mol.molecular_charge == 0
    assert mol.molecular_multiplicity == 1
    assert np.sum(mol.real) == mol.geometry.shape[0]
    assert np.allclose(mol.fragments, [[0, 1, 2], [3, 4, 5]])
    assert np.allclose(mol.fragment_charges, [0, 0])
    assert np.allclose(mol.fragment_multiplicities, [1, 1])
    assert hasattr(mol, "provenance")
    assert np.allclose(
        mol.geometry,
        [[2.81211080, 0.1255717, 0.], [3.48216664, -1.55439981, 0.], [1.00578203, -0.1092573, 0.],
         [-2.6821528, -0.12325075, 0.], [-3.27523824, 0.81341093, 1.43347255], [-3.27523824, 0.81341093, -1.43347255]])
    assert mol.get_hash() == "3c4b98f515d64d1adc1648fe1fe1d6789e978d34"


def test_water_minima_fragment():

    mol = water_dimer_minima.copy()
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "5f31757232a9a594c46073082534ca8a6806d367"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"

    frag_0_1 = mol.get_fragment(0, 1)
    frag_1_0 = mol.get_fragment(1, 0)

    assert np.array_equal(mol.symbols[:3], frag_0.symbols)
    assert np.allclose(mol.masses[:3], frag_0.masses)

    assert np.array_equal(mol.symbols, frag_0_1.symbols)
    assert np.allclose(mol.geometry, frag_0_1.geometry)

    assert np.array_equal(np.hstack((mol.symbols[3:], mol.symbols[:3])), frag_1_0.symbols)
    assert np.allclose(np.hstack((mol.masses[3:], mol.masses[:3])), frag_1_0.masses)


def test_pretty_print():

    mol = water_dimer_minima.copy()
    assert isinstance(mol.pretty_print(), str)


def test_to_string():

    mol = water_dimer_minima.copy()
    assert isinstance(mol.to_string("psi4"), str)


@pytest.mark.parametrize("dtype, filext", [
    ("json", "json"),
    ("xyz", "xyz"),
    ("numpy", "npy"),
    pytest.param("msgpack", "msgpack", marks=using_msgpack),
])
def test_to_from_file_simple(tmp_path, dtype, filext):

    benchmol = Molecule.from_data("""
    O 0 0 0
    H 0 1.5 0
    H 0 0 1.5
    """)

    p = tmp_path / ("water." + filext)
    benchmol.to_file(p)

    mol = Molecule.from_file(p)

    assert mol.compare(benchmol)


@pytest.mark.parametrize("dtype", ["json", "psi4"])
def test_to_from_file_complex(tmp_path, dtype):

    p = tmp_path / ("water." + dtype)
    water_dimer_minima.to_file(p)

    mol = Molecule.from_file(p)
    assert mol.compare(water_dimer_minima)


def test_water_orient():
    # These are identical molecules, should find the correct results
    mol = Molecule.from_data("""
        O  -1.551007  -0.114520   0.000000
        H  -1.934259   0.762503   0.000000
        H  -0.599677   0.040712   0.000000
        --
        O  -0.114520  -1.551007  10.000000
        H   0.762503  -1.934259  10.000000
        H   0.040712  -0.599677  10.000000
        """)

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=True)
    assert frag_0_1.get_hash() == frag_1_0.get_hash()

    # Fragments not reordered, should be different molecules.
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=False)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=False)
    assert frag_0_1.get_hash() != frag_1_0.get_hash()

    # These are identical molecules, but should be different with ghost
    mol = Molecule.from_data("""
        O  -1.551007  -0.114520   0.000000
        H  -1.934259   0.762503   0.000000
        H  -0.599677   0.040712   0.000000
        --
        O  -11.551007  -0.114520   0.000000
        H  -11.934259   0.762503   0.000000
        H  -10.599677   0.040712   0.000000
        """,
                             dtype="psi4",
                             orient=True)

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.molecular_multiplicity == 1
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True)

    # Ghost fragments should prevent overlap
    assert frag_0_1.molecular_multiplicity == 1
    assert frag_0_1.get_hash() != frag_1_0.get_hash()


def test_molecule_errors_extra():
    data = water_dimer_minima.dict(skip_defaults=True)
    data["whatever"] = 5
    with pytest.raises(Exception):
        Molecule(**data, validate=False)


def test_molecule_errors_connectivity():
    data = water_molecule.dict()
    data["connectivity"] = [(-1, 5, 5)]
    with pytest.raises(Exception):
        Molecule(**data)


def test_molecule_errors_shape():
    data = water_molecule.dict()
    data["geometry"] = list(range(8))
    with pytest.raises(Exception):
        Molecule(**data)


def test_molecule_json_serialization():
    assert isinstance(water_dimer_minima.json(), str)

    assert isinstance(water_dimer_minima.dict(encoding="json")["geometry"], list)

    assert water_dimer_minima.compare(Molecule.from_data(water_dimer_minima.json(), dtype="json"))


@pytest.mark.parametrize("encoding", serialize_extensions)
def test_molecule_serialization(encoding):
    blob = water_dimer_minima.serialize(encoding)
    assert water_dimer_minima.compare(Molecule.parse_raw(blob, encoding=encoding))


def test_charged_fragment():
    mol = Molecule(symbols=["Li", "Li"],
                   geometry=[0, 0, 0, 0, 0, 5],
                   fragment_charges=[0.0, 0.0],
                   fragment_multiplicities=[2, 2],
                   fragments=[[0], [1]])
    assert mol.molecular_multiplicity == 3
    assert mol.molecular_charge == 0
    f1 = mol.get_fragment(0)
    assert f1.molecular_multiplicity == 2
    assert f1.fragment_multiplicities == [2]
    assert pytest.approx(f1.molecular_charge) == 0
    assert pytest.approx(f1.fragment_charges) == [0]


# someday, when we can have noncontig frags
#    mol = Molecule(**{
#        'fragments': [[1, 5, 6], [0], [2, 3, 4]],
#        'symbols': ["he", "o", "o", "h", "h", "h", "h"],
#        # same geom as test_water_orient but with He at origin
#        'geometry': [
#        0.0, 0.0, 0.0,
#        -1.551007,  -0.114520,   0.000000,
#        -0.114520,  -1.551007,  10.000000,
#         0.762503,  -1.934259,  10.000000,
#         0.040712,  -0.599677,  10.000000,
#        -1.934259,   0.762503,   0.000000,
#        -0.599677,   0.040712,   0.000000],
#    })


@pytest.mark.parametrize(
    "group_fragments, orient",
    [
        (True, True),  # original
        (False, False),  # Psi4-like
    ])
def test_get_fragment(group_fragments, orient):
    mol = Molecule(
        **{
            'fragments': [[0], [1, 2, 3], [4, 5, 6]],
            'symbols': ["he", "o", "h", "h", "o", "h", "h"],
            # same geom as test_water_orient but with He at origin
            'geometry':
            np.array([
                0.0, 0.0, 0.0, -1.551007, -0.114520, 0.000000, -1.934259, 0.762503, 0.000000, -0.599677, 0.040712,
                0.000000, -0.114520, -1.551007, 10.000000, 0.762503, -1.934259, 10.000000, 0.040712, -0.599677,
                10.000000
            ]) / qcel.constants.bohr2angstroms,
        })

    assert mol.nelectrons() == 22
    assert compare_values(32.25894779318589, mol.nuclear_repulsion_energy(), atol=1.e-5)
    assert mol.symbols[0] == "He"

    monomers_nelectrons = [2, 10, 10]
    monomers_nre = [0.0, 9.163830150548483, 9.163830150548483]
    monomers = [mol.get_fragment(ifr, group_fragments=group_fragments, orient=orient) for ifr in range(3)]
    for fr in range(3):
        assert monomers[fr].nelectrons() == monomers_nelectrons[fr]
        assert compare_values(monomers[fr].nuclear_repulsion_energy(), monomers_nre[fr], 'monomer nre', atol=1.e-5)

    idimers = [(0, 1), (0, 2), (1, 2), (1, 0), (2, 0), (2, 1)]
    dimers_nelectrons = [12, 12, 20, 12, 12, 20]
    dimers_nre = [16.8777971, 10.2097206, 23.4990904, 16.8777971, 10.2097206, 23.4990904]
    dimers = [mol.get_fragment(rl, group_fragments=group_fragments, orient=orient) for rl in idimers]
    for ifr in range(len(idimers)):
        # print('dd', ifr, idimers[ifr], dimers[ifr].nuclear_repulsion_energy(), dimers[ifr].get_hash())
        assert dimers[ifr].nelectrons() == dimers_nelectrons[ifr], 'dimer nelec'
        assert compare_values(dimers[ifr].nuclear_repulsion_energy(), dimers_nre[ifr], 'dimer nre', atol=1.e-5)
    if group_fragments and orient:
        assert dimers[0].get_hash() != dimers[3].get_hash()  # atoms out of order
        assert dimers[1].get_hash() != dimers[4].get_hash()  # atoms out of order
        assert dimers[2].get_hash() == dimers[5].get_hash()
    elif not group_fragments and not orient:
        assert dimers[0].get_hash() == dimers[3].get_hash()
        assert dimers[1].get_hash() == dimers[4].get_hash()
        assert dimers[2].get_hash() == dimers[5].get_hash()
    else:
        assert 0  # lgtm: [py/unreachable-statement]

    ghdimers_nelectrons = [2, 2, 10, 10, 10, 10]
    ghdimers_nre = [0.0, 0.0, 9.163830150548483, 9.163830150548483, 9.163830150548483, 9.163830150548483]
    ghdimers = [mol.get_fragment(rl, gh, group_fragments=group_fragments, orient=orient) for rl, gh in idimers]
    for ifr in range(len(idimers)):
        # print('gh', ifr, idimers[ifr], ghdimers[ifr].nuclear_repulsion_energy(), ghdimers[ifr].get_hash())
        assert ghdimers[ifr].nelectrons() == ghdimers_nelectrons[ifr], 'gh dimer nelec'
        assert compare_values(ghdimers[ifr].nuclear_repulsion_energy(), ghdimers_nre[ifr], 'gh dimer nre', atol=1.e-5)

    if group_fragments and orient:
        assert ghdimers[0].get_hash() != ghdimers[3].get_hash()  # diff atoms ghosted
        assert ghdimers[1].get_hash() != ghdimers[4].get_hash()  # diff atoms ghosted
        assert ghdimers[2].get_hash() == ghdimers[5].get_hash()
    elif not group_fragments and not orient:
        assert ghdimers[0].get_hash() != ghdimers[3].get_hash()  # diff atoms ghosted
        assert ghdimers[1].get_hash() != ghdimers[4].get_hash()  # diff atoms ghosted
        assert ghdimers[2].get_hash() != ghdimers[5].get_hash()  # real pattern different
        assert not np.allclose(ghdimers[2].real, ghdimers[5].real)
    else:
        assert 0  # lgtm: [py/unreachable-statement]


def test_molecule_repeated_hashing():

    mol = Molecule(**{
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.7317,  1.2909,  1.037100000000001,
             1.3156, -0.0074, -0.2807,
            -1.3143,  0.0084, -0.2741,
            -1.7241, -1.3079,  1.0277
        ]
    })  # yapf: disable

    h1 = mol.get_hash()
    assert mol.get_molecular_formula() == "H2O2"

    mol2 = Molecule(orient=False, **mol.dict())
    assert h1 == mol2.get_hash()

    mol3 = Molecule(orient=False, **mol2.dict())
    assert h1 == mol3.get_hash()


@pytest.mark.parametrize("measure,result", [
    ([0, 1], 1.8086677572537304),
    ([0, 1, 2], 37.98890673587713),
    ([0, 1, 2, 3], 180.0),
    ([[0, 1, 2, 3]], [180.0]),
    ([[1, 3], [3, 1], [1, 2, 3]], [6.3282716, 6.3282716, 149.51606694803903]),
])
def test_measurements(measure, result):

    Molecule(**{
        'symbols': ['H', 'O', 'O', 'H'],
        'geometry': [
             1.7317,  1.2909,  1.0371,
             1.3156, -0.0074, -0.2807,
            -1.3143,  0.0084, -0.2741,
            -1.7241, -1.3079,  1.0277
        ]
    })  # yapf: disable

    assert pytest.approx(water_dimer_minima.measure(measure)) == result


@pytest.mark.parametrize("f1c,f1m,f2c,f2m,tc,tm", [
    (0, 2, 0, 2, 0.0, 3),
    (0, 4, 0, 2, 0.0, 5),
    (1, 1, 1, 1, 2.0, 1),
    (1, 1, 0, 2, 1.0, 2),
    (0, 2, 1, 1, 1.0, 2),
    (-1, 1, 1, 1, 0.0, 1),
])
def test_fragment_charge_configurations(f1c, f1m, f2c, f2m, tc, tm):

    mol = Molecule.from_data("""
    {f1c} {f1m}
    Li 0 0 0
    --
    {f2c} {f2m}
    Li 0 0 5
    """.format(f1c=f1c, f1m=f1m, f2c=f2c, f2m=f2m))

    assert pytest.approx(mol.molecular_charge) == tc
    assert mol.molecular_multiplicity == tm

    # Test fragment1
    assert pytest.approx(mol.get_fragment(0).molecular_charge) == f1c
    assert mol.get_fragment(0).molecular_multiplicity == f1m

    assert pytest.approx(mol.get_fragment(0, 1).molecular_charge) == f1c
    assert mol.get_fragment(0, 1).molecular_multiplicity == f1m

    # Test fragment2
    assert pytest.approx(mol.get_fragment(1).molecular_charge) == f2c
    assert mol.get_fragment(1).molecular_multiplicity == f2m

    assert pytest.approx(mol.get_fragment([1], 0).molecular_charge) == f2c
    assert mol.get_fragment(1, [0]).molecular_multiplicity == f2m


def test_nuclearrepulsionenergy_nelectrons():

    mol = Molecule.from_data("""
    0 1
    --
    O          0.75119       -0.61395        0.00271
    H          1.70471       -0.34686        0.00009
    --
    1 1
    N         -2.77793        0.00179       -0.00054
    H         -2.10136        0.51768        0.60424
    H         -3.45559       -0.51904        0.60067
    H         -2.26004       -0.67356       -0.60592
    H         -3.29652        0.68076       -0.60124
    units ang
    """)

    assert compare_values(34.60370459, mol.nuclear_repulsion_energy(), 'D', atol=1.e-5)
    assert compare_values(4.275210518, mol.nuclear_repulsion_energy(ifr=0), 'M1', atol=1.e-5)
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(ifr=1), 'M2', atol=1.e-5)

    assert compare(20, mol.nelectrons(), 'D')
    assert compare(10, mol.nelectrons(ifr=0), 'M1')
    assert compare(10, mol.nelectrons(ifr=1), 'M2')

    mol = mol.get_fragment([1], 0, group_fragments=False)
    # Notice the 0th/1st fragments change if default group_fragments=True.
    ifr0 = 0
    ifr1 = 1
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(), 'D', atol=1.e-5)
    assert compare_values(0.0, mol.nuclear_repulsion_energy(ifr=ifr0), 'M1', atol=1.e-5)
    assert compare_values(16.04859029, mol.nuclear_repulsion_energy(ifr=ifr1), 'M2', atol=1.e-5)

    assert compare(10, mol.nelectrons(), 'D')
    assert compare(0, mol.nelectrons(ifr=ifr0), 'M1')
    assert compare(10, mol.nelectrons(ifr=ifr1), 'M2')


@using_py3dmol
def test_show():

    water_dimer_minima.show()



import os
from decimal import Decimal

import pytest

import qcelemental


@pytest.mark.parametrize("inp,expected", [
    ("He", "He"),
    ("He", "He"),
    ("He4", "He"),
    ("he", "He"),
    ("2", "He"),
    (2, "He"),
    (2.0, "He"),
])
def test_id_resolution(inp, expected):
    assert qcelemental.periodictable._resolve_atom_to_key('He') == 'He'


@pytest.mark.parametrize("inp", ["He100", '-1', -1, -1.0, 'cat', 200])
def test_id_resolution_error(inp):
    with pytest.raises(qcelemental.exceptions.NotAnElementError):
        qcelemental.periodictable._resolve_atom_to_key(inp)


# TODO test ghost


def test_to_mass_krypton_decimal():
    assert qcelemental.periodictable.to_mass('kr', return_decimal=True) == Decimal('83.9114977282')


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("KRYPTON", 83.9114977282),
        ("kr", 83.9114977282),
        ("kr84", 83.9114977282),
        (36, 83.9114977282),

        # Kr 86
        ("kr86", 85.9106106269),

        # Helium
        ("D", 2.01410177812),
        ("h2", 2.01410177812),
    ])
def test_to_mass(inp, expected):
    assert qcelemental.periodictable.to_mass(inp) == pytest.approx(expected, 1.e-9)


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", 84),
        ("KRYPTON", 84),
        ("kr84", 84),
        (36, 84),

        # Kr 86
        ("kr86", 86),

        # Helium
        ("D", 2),
        ("h2", 2),
    ])
def test_to_mass_number(inp, expected):
    assert qcelemental.periodictable.to_A(inp) == expected
    assert qcelemental.periodictable.to_mass_number(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", 36),
        ("KRYPTON", 36),
        ("kr84", 36),
        (36, 36),

        # Kr 86
        ("kr86", 36),

        # Deuterium
        ("D", 1),
        ("h2", 1),
    ])
def test_to_atomic_number(inp, expected):
    assert qcelemental.periodictable.to_Z(inp) == expected
    assert qcelemental.periodictable.to_atomic_number(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Kr"),
        ("KRYPTON", "Kr"),
        ("kr84", "Kr"),
        (36, "Kr"),

        # Kr 86
        ("kr86", "Kr"),

        # Deuterium
        ("D", "H"),
        ("h2", "H"),
    ])
def test_to_symbol(inp, expected):
    assert qcelemental.periodictable.to_E(inp) == expected
    assert qcelemental.periodictable.to_symbol(inp) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("kr", "Krypton"),
        ("KRYPTON", "Krypton"),
        ("kr84", "Krypton"),
        (36, "Krypton"),

        # Kr 86
        ("kr86", "Krypton"),

        # Deuterium
        ("D", "Hydrogen"),
        ("h2", "Hydrogen"),
    ])
def test_to_element(inp, expected):
    assert qcelemental.periodictable.to_element(inp) == expected
    assert qcelemental.periodictable.to_name(inp) == expected


@pytest.mark.parametrize("inp,expected", [('HE', 1), ('carbon', 2), ('cl35', 3), (36, 4), (37.0, 5), ('mercury', 6),
                                          ('Ts', 7)])
def test_to_period(inp, expected):
    assert qcelemental.periodictable.to_period(inp) == expected


@pytest.mark.parametrize("inp,expected", [
    ('D', 1),
    ('Ba', 2),
    ('yttrium', 3),
    ('hf179', 4),
    (23, 5),
    ('w', 6),
    (43.0, 7),
    (26.0, 8),
    ('meitnerium', 9),
    ('pt', 10),
    ('gold', 11),
    ('cd', 12),
    (5.0, 13),
    (14, 14),
    ('phosphorus', 15),
    ('SE', 16),
    ('i', 17),
    (54, 18),
    ('La', None),
    (89.0, None),
    ('lutetium', None),
    (103.0, None),
    ('u238', None),
])
def test_to_group(inp, expected):
    assert qcelemental.periodictable.to_group(inp) == expected


def test_c_header():
    qcelemental.periodictable.write_c_header("header.h")
    os.remove("header.h")


@pytest.mark.xfail(True, reason='comparison data not available for installed repository', run=True, strict=False)
def test_periodic_table_comparison():
    qcelemental.periodictable.run_comparison()



import copy
from typing import Dict

import numpy as np
import pytest

import qcelemental
from qcelemental.testing import compare_molrecs

_string_prov_stamp = {'creator': 'QCElemental', 'version': '1.0', 'routine': 'qcelemental.molparse.from_string'}
_schema_prov_stamp = {'creator': 'QCElemental', 'version': '1.0', 'routine': 'qcelemental.molparse.from_schema'}

subject14 = """0 3\n--\nHe 0 0 -5\n--\n@He 0 0 5\nunits au"""

schema14_1 = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule": {
        "validated": True,
        "geometry": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
        "symbols": ["He", "He"],
        "atomic_numbers": [2, 2],
        "mass_numbers": [4, 4],
        "atom_labels": ['', ''],
        'fragments': [[0], [1]],
        'fragment_charges': [0.0, 0.0],
        'fragment_multiplicities': [3, 1],
        'masses': [4.00260325413, 4.00260325413],
        'name': 'He2',
        'fix_com': False,
        'fix_orientation': False,
        'molecular_charge': 0.0,
        "molecular_multiplicity": 3,
        "real": [True, False]
    }
}

schema14_2 = {
    'schema_name': 'qcschema_molecule',
    'schema_version': 2,
}
schema14_2.update(schema14_1['molecule'])

schema14_psi4 = {
    "geom": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "elem": ["He", "He"],
    'elea': [4, 4],
    'elez': [2, 2],
    'fragment_charges': [0.0, 0.0],
    'fragment_multiplicities': [3, 1],
    'mass': [4.00260325413, 4.00260325413],
    'name': 'He2',
    'fix_com': False,
    'fix_orientation': False,
    'molecular_charge': 0.0,
    "molecular_multiplicity": 3,
    'units': 'Bohr',
    'fragment_separators': [1],
    'elbl': ['', ''],
    "real": [True, False]
}


def test_1_14a():
    fullans = copy.deepcopy(schema14_1)
    fullans['molecule']['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=1)
    assert compare_molrecs(fullans['molecule'], kmol['molecule'])

    fullans = copy.deepcopy(schema14_psi4)
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_2_14b():
    fullans = copy.deepcopy(schema14_2)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=2)
    assert compare_molrecs(fullans, kmol)

    fullans = copy.deepcopy(schema14_psi4)
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_psi4_14c():
    fullans = copy.deepcopy(schema14_psi4)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject14)
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype='psi4')
    assert compare_molrecs(fullans, kmol)


def test_dtype_error():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype='xkcd927')

    assert "Schema dtype not understood" in str(e.value)


@pytest.mark.parametrize("dtype", [
    1,
    2,
])
def test_qcschema_ang_error(dtype):

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype=dtype, units='Angstrom')

    assert f"QC_JSON_Schema {dtype} allows only 'Bohr' coordinates" in str(e.value)


def test_psi4_nm_error():

    final = qcelemental.molparse.from_string(subject14)
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.to_schema(final['qm'], dtype='psi4', units='nm')

    assert "Psi4 Schema psi4 allows only 'Bohr'/'Angstrom' coordinates, not nm" in str(e.value)


twobohrinang = 2.0 * qcelemental.constants.conversion_factor("bohr", "angstrom")
subject15 = """symmetry cS\nH 0 0 {twobohrinang}\nO 0 0 0\n2H_deut {twobohrinang} 0 0\nno_com\nno_reorient""".format(
    twobohrinang=twobohrinang)

schema15_1: Dict = {
    "schema_name": "qc_schema_input",
    "schema_version": 1,
    "molecule": {
        "validated": True,
        "geometry": [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0],
        "symbols": ["H", "O", "H"],
        "atomic_numbers": [1, 8, 1],
        "mass_numbers": [1, 16, 2],
        "atom_labels": ['', '', '_deut'],
        'fragments': [[0, 1, 2]],
        'fragment_charges': [0.0],
        'fragment_multiplicities': [1],
        'masses': [1.00782503223, 15.99491461957, 2.01410177812],
        'name': 'H2O',
        'comment': 'I has a comment',
        'fix_com': True,
        'fix_orientation': True,
        'fix_symmetry': 'cs',
        'molecular_charge': 0.0,
        "molecular_multiplicity": 1,
        "real": [True, True, True]
    }
}

schema15_2 = {'schema_name': 'qcschema_molecule', 'schema_version': 2}
schema15_2.update(schema15_1['molecule'])

schema15_psi4 = {
    "geom": [0.0, 0.0, twobohrinang, 0.0, 0.0, 0.0, twobohrinang, 0.0, 0.0],
    "elem": ["H", "O", "H"],
    "elez": [1, 8, 1],
    "elea": [1, 16, 2],
    "elbl": ['', '', '_deut'],
    'fragment_charges': [0.0],
    'fragment_multiplicities': [1],
    'mass': [1.00782503223, 15.99491461957, 2.01410177812],
    'name': 'H2O',
    'comment': 'I has a comment',
    'fix_com': True,
    'fix_orientation': True,
    'fix_symmetry': 'cs',
    'molecular_charge': 0.0,
    "molecular_multiplicity": 1,
    'units': 'Angstrom',
    'fragment_separators': [],
    "real": [True, True, True]
}


def test_1_15a():
    fullans = copy.deepcopy(schema15_1)
    fullans['molecule']['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final['qm']['comment'] = 'I has a comment'
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=1)
    assert compare_molrecs(fullans['molecule'], kmol['molecule'])

    fullans = copy.deepcopy(schema15_psi4)
    fullans['units'] = 'Bohr'
    fullans['geom'] = [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0]
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_2_15b():
    fullans = copy.deepcopy(schema15_2)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final['qm']['comment'] = 'I has a comment'
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype=2)
    assert compare_molrecs(fullans, kmol)

    fullans = copy.deepcopy(schema15_psi4)
    fullans['units'] = 'Bohr'
    fullans['geom'] = [0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0]
    fullans['provenance'] = _schema_prov_stamp

    molrec = qcelemental.molparse.from_schema(kmol)
    molrec = qcelemental.util.unnp(molrec)
    assert compare_molrecs(fullans, molrec)


def test_psi4_15c():
    fullans = copy.deepcopy(schema15_psi4)
    fullans['provenance'] = _string_prov_stamp

    final = qcelemental.molparse.from_string(subject15)
    final['qm']['comment'] = 'I has a comment'
    kmol = qcelemental.molparse.to_schema(final['qm'], dtype='psi4', units='Angstrom')
    assert compare_molrecs(fullans, kmol)


schema16_1 = {
    'schema_name': 'qc_schema_input',
    'schema_version': 1,
    'molecule': {
        'validated': True,
        'geometry': [2., 2., 3.],
        'symbols': ['C'],
        'masses': [12.0],
        'atom_labels': [''],
        'atomic_numbers': [6],
        'mass_numbers': [12],
        'real': [True],
        'name': 'C',
        'molecular_charge': 0.,
        'molecular_multiplicity': 1,
        'fragments': [[0]],
        'fragment_charges': [0.],
        'fragment_multiplicities': [1],
        'fix_com': False,
        'fix_orientation': False,
        'provenance': {
            'creator': 'Mystery Program',
            'version': '2018.3',
            'routine': 'molecule builder',
        },
        'connectivity': [(0, 0, 0.0)],
    },
}

schema16_2 = {'schema_name': 'qcschema_molecule', 'schema_version': 2}
schema16_2.update(schema16_1['molecule'])

schema16_psi4 = {
    'units': 'Bohr',
    'geom': np.array([2., 2., 3.]),
    'elem': np.array(['C']),
    'mass': np.array([12.0]),
    'elbl': np.array(['']),
    'elez': np.array([6]),
    'elea': np.array([12]),
    'real': np.array([True]),
    'name': 'C',
    'molecular_charge': 0.,
    'molecular_multiplicity': 1,
    'fragment_separators': [],
    'fragment_charges': [0.],
    'fragment_multiplicities': [1],
    'fix_com': False,
    'fix_orientation': False,
    'provenance': {
        'creator': 'Mystery Program',
        'version': '2018.3',
        'routine': 'molecule builder',
    },
    'connectivity': [(0, 0, 0.0)],
}


def test_froto_1_16a():
    basic = {
        'schema_name': 'qc_schema_output',
        'schema_version': 1,
        'molecule': {
            'geometry': [2, 2, 3],
            'symbols': ['C'],
            'connectivity': [(0.0, -0.0, 0)],
            'provenance': {
                'creator': 'Mystery Program',
                'version': '2018.3',
                'routine': 'molecule builder',
            },
        },
    }

    fullans = copy.deepcopy(schema16_1)
    fullans['molecule']['provenance'] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.to_schema(qcelemental.molparse.from_schema(basic), dtype=1)
    assert compare_molrecs(fullans['molecule'], roundtrip['molecule'])


def test_froto_2_16a():
    basic = {
        'schema_name': 'qcschema_molecule',
        'schema_version': 2,
        'geometry': [2, 2, 3],
        'symbols': ['C'],
        'connectivity': [(0.0, -0.0, 0)],
        'provenance': {
            'creator': 'Mystery Program',
            'version': '2018.3',
            'routine': 'molecule builder',
        },
    }

    fullans = copy.deepcopy(schema16_2)
    fullans['provenance'] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.to_schema(qcelemental.molparse.from_schema(basic), dtype=2)
    assert compare_molrecs(fullans, roundtrip)


@pytest.mark.parametrize("dtype", [
    1,
    2,
])
def test_tofro_16b(dtype):

    fullans = copy.deepcopy(schema16_psi4)
    fullans['provenance'] = _schema_prov_stamp

    roundtrip = qcelemental.molparse.from_schema(qcelemental.molparse.to_schema(schema16_psi4, dtype=dtype))
    assert compare_molrecs(fullans, roundtrip)



import copy

import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.testing import compare_molrecs

_schema_prov_stamp = {'creator': 'QCElemental', 'version': '1.0', 'routine': 'qcelemental.molparse.from_schema'}


@pytest.mark.parametrize("inp,expected", [
    ({
        'frag_pattern': [[0], [1]],
        'geom': [0., 0., 0., 1., 0., 0.],
        'elbl': ['O', 'H']
    }, {
        'fragment_separators': np.array([1]),
        'geom': np.array([0., 0., 0., 1., 0., 0.]),
        'elbl': np.array(['O', 'H'])
    }),
    ({
        'frag_pattern': [[2, 0], [1]],
        'geom': np.array([[0., 0., 1.], [0., 0., 2.], [0., 0., 0.]]),
        'elem': np.array(['Li', 'H', 'He'])
    }, {
        'fragment_separators': np.array([2]),
        'geom': np.array([0., 0., 0., 0., 0., 1., 0., 0., 2.]),
        'elem': np.array(['He', 'Li', 'H'])
    }),
    ({
        'frag_pattern': [[2, 0], [1]],
        'elez': [3, 1, 2]
    }, {
        'fragment_separators': np.array([2]),
        'elez': np.array([2, 3, 1])
    }),
])
def test_contiguize_from_fragment_pattern(inp, expected):
    ans = qcel.molparse.contiguize_from_fragment_pattern(**inp)

    # compare_molrecs instead of compare_dicts handles some fragment_separators types issues
    assert compare_molrecs(expected, ans, atol=1.e-6)


@pytest.mark.parametrize("inp,expected", [
    ({
        'frag_pattern': [[2, 0], [1, 3]],
        'geom': np.array([[0., 0., 1.], [0., 0., 2.], [0., 0., 0.]]),
        'elem': np.array(['Li', 'H', 'He'])
    }, 'dropped atoms'),
    ({
        'frag_pattern': [[2, 0], [1, 4]]
    }, 'Fragmentation pattern skips atoms'),
    ({
        'frag_pattern': [[2, 0], [1, 3]],
        'elem': np.array(['U', 'Li', 'H', 'He']),
        'elbl': np.array(['Li', 'H', 'He'])
    }, 'wrong number of atoms in array'),
    ({
        'frag_pattern': [[2, 0], [1]],
        'elez': [3, 1, 2],
        'throw_reorder': True
    }, 'reorder atoms to accommodate non-contiguous fragments'),
])
def test_contiguize_from_fragment_pattern_error(inp, expected):
    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.contiguize_from_fragment_pattern(**inp)

    assert expected in str(e.value)


schema14_1 = {
    "geometry": [0.0, 0.0, -5.0, 0.0, 0.0, 5.0],
    "symbols": ["He", "He"],
    'fragments': [[0], [1]],
    'fragment_charges': [0.0, 0.0],
    'fragment_multiplicities': [3, 1],
    'masses': [4.00260325413, 4.00260325413],
    'name': 'He2',
    'fix_com': False,
    'fix_orientation': False,
    'molecular_charge': 0.0,
    "molecular_multiplicity": 3,
    "real": [True, False]
}

schema14_psi4_np = {
    "geom": np.array([0.0, 0.0, -5.0, 0.0, 0.0, 5.0]),
    "elem": np.array(["He", "He"]),
    'elea': np.array([4, 4]),
    'elez': np.array([2, 2]),
    'fragment_charges': [0.0, 0.0],
    'fragment_multiplicities': [3, 1],
    'mass': np.array([4.00260325413, 4.00260325413]),
    'name': 'He2',
    'fix_com': False,
    'fix_orientation': False,
    'molecular_charge': 0.0,
    "molecular_multiplicity": 3,
    'units': 'Bohr',
    'fragment_separators': [1],
    'elbl': np.array(['', '']),
    "real": np.array([True, False]),
    "provenance": _schema_prov_stamp,
}


def test_from_schema_1_14e():
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_1p5_14e():
    # this is the oddball case where passing code has internally a dtype=2 molecule
    #   but it's still passing the outer data structure
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}
    schema['molecule'].update({"schema_name": "qcschema_molecule", "schema_version": 2})

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_2_14e():
    schema = copy.deepcopy(schema14_1)
    schema.update({"schema_name": "qcschema_molecule", "schema_version": 2})

    ans = qcel.molparse.from_schema(schema)
    assert compare_molrecs(schema14_psi4_np, ans, 4)


def test_from_schema_error_f():
    schema = {"schema_name": "private_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}

    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.from_schema(schema)

    assert 'Schema not recognized' in str(e.value)


def test_from_schema_1_nfr_error_14g():
    schema = {"schema_name": "qc_schema", "schema_version": 1, "molecule": copy.deepcopy(schema14_1)}
    schema['molecule'].pop('fragments')

    with pytest.raises(qcel.ValidationError) as e:
        qcel.molparse.from_schema(schema)

    assert 'Dimension mismatch among fragment quantities: sep + 1 (1), chg (2), and mult(2)' in str(e.value)



import pytest

import qcelemental
from qcelemental.testing import compare, compare_values


@pytest.mark.parametrize("inp,expected", [
    ('@ca_miNe', {'E': 'ca', 'Z': None, 'user': '_miNe', 'A': None, 'real': False, 'mass': None}),
    ('Gh(Ca_mine)', {'E': 'Ca', 'Z': None, 'user': '_mine', 'A': None, 'real': False, 'mass': None}),
    ('@Ca_mine@1.07', {'E': 'Ca', 'Z': None, 'user': '_mine', 'A': None, 'real': False, 'mass': 1.07}),
    ('Gh(cA_MINE@1.07)', {'E': 'cA', 'Z': None, 'user': '_MINE', 'A': None, 'real': False, 'mass': 1.07}),
    ('@40Ca_mine@1.07', {'E': 'Ca', 'Z': None, 'user': '_mine', 'A': 40, 'real': False, 'mass': 1.07}),
    ('Gh(40Ca_mine@1.07)', {'E': 'Ca', 'Z': None, 'user': '_mine', 'A': 40, 'real': False, 'mass': 1.07}),
    ('444lu333@4.0', {'E': 'lu', 'Z': None, 'user': '333', 'A': 444, 'real': True, 'mass': 4.0}),
    ('@444lu333@4.4', {'E': 'lu', 'Z': None, 'user': '333', 'A': 444, 'real': False, 'mass': 4.4}),
    ('8i', {'E': 'i', 'Z': None, 'user': None, 'A': 8, 'real': True, 'mass': None}),
    ('53_mI4', {'Z': 53, 'E': None, 'user': '_mI4', 'A': None, 'real': True, 'mass': None}),
    ('@5_MINEs3@4.4', {'Z': 5, 'E': None, 'user': '_MINEs3', 'A': None, 'real': False, 'mass': 4.4}),
    ('Gh(555_mines3@0.1)', {'Z': 555, 'E': None, 'user': '_mines3', 'A': None, 'real': False, 'mass': 0.1}),
])  # yapf: disable
def test_parse_nucleus_label(inp, expected):
    lbl_A, lbl_Z, lbl_E, lbl_mass, lbl_real, lbl_user = qcelemental.molparse.nucleus.parse_nucleus_label(inp)

    assert compare(expected['real'], lbl_real, inp + " real")
    assert compare(expected['A'], lbl_A, inp + " A")
    assert compare(expected['Z'], lbl_Z, inp + " Z")
    assert compare(expected['E'], lbl_E, inp + " symbol")
    assert compare(expected['user'], lbl_user, inp + " user")
    assert compare_values(expected['mass'], lbl_mass, inp + " mass", passnone=True, atol=1.e-6)


@pytest.mark.parametrize("inp", [
    '1L1A1B1',
])
def test_parse_nucleus_label_error(inp):
    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.nucleus.parse_nucleus_label(inp)



import json
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import numpy as np
from pydantic import BaseModel, BaseSettings
from pydantic.main import MetaModel

from qcelemental.testing import compare_recursive
from qcelemental.util import deserialize, serialize
from qcelemental.util.autodocs import AutoPydanticDocGenerator


class PydanticAutodocMeta(MetaModel):
    def __new__(mcs, name, bases, namespace):

        ret = super().__new__(mcs, name, bases, namespace)
        ret.__doc__ = AutoPydanticDocGenerator(ret, always_apply=True)
        return ret


class ProtoModel(BaseModel, metaclass=PydanticAutodocMeta):
    class Config:
        allow_mutation = False
        extra = "forbid"
        json_encoders = {np.ndarray: lambda v: v.flatten().tolist()}
        serialize_default_excludes: Set = set()
        serialize_skip_defaults = False
        force_skip_defaults = False
        canonical_repr = False

    @classmethod
    def parse_raw(cls, data: Union[bytes, str], *, encoding: str = None) -> 'ProtoModel':  # type: ignore
        """
        Parses raw string or bytes into a Model object.

        Parameters
        ----------
        data : Union[bytes, str]
            A serialized data blob to be deserialized into a Model.
        encoding : str, optional
            The type of the serialized array, available types are: {'json', 'json-ext', 'msgpack-ext', 'pickle'}

        Returns
        -------
        Model
            The requested model from a serialized format.
        """

        if encoding is None:
            if isinstance(data, str):
                encoding = "json"
            elif isinstance(data, bytes):
                encoding = "msgpack-ext"
            else:
                raise TypeError("Input is neither str nor bytes, please specify an encoding.")

        if encoding.endswith(('json', 'javascript', 'pickle')):
            return super().parse_raw(data, content_type=encoding)
        elif encoding in ["msgpack-ext", "json-ext"]:
            obj = deserialize(data, encoding)
        else:
            raise TypeError(f"Content type '{encoding}' not understood.")

        return cls.parse_obj(obj)

    @classmethod
    def parse_file(cls, path: Union[str, Path], *, encoding: str = None) -> 'ProtoModel':  # type: ignore
        """Parses a file into a Model object.

        Parameters
        ----------
        path : Union[str, Path]
            The path to the file.
        encoding : str, optional
            The type of the files, available types are: {'json', 'msgpack', 'pickle'}. Attempts to
            automatically infer the file type from the file extension if None.

        Returns
        -------
        Model
            The requested model from a serialized format.

        """
        path = Path(path)
        if encoding is None:
            if path.suffix in [".json", ".js"]:
                encoding = "json"
            elif path.suffix in [".msgpack"]:
                encoding = "msgpack-ext"
            elif path.suffix in [".pickle"]:
                encoding = "pickle"
            else:
                raise TypeError("Could not infer `encoding`, please provide a `encoding` for this file.")

        return cls.parse_raw(path.read_bytes(), encoding=encoding)

    def dict(self, **kwargs) -> Dict[str, Any]:
        encoding = kwargs.pop("encoding", None)

        kwargs["exclude"] = (
            (kwargs.get("exclude", None) or set()) | self.__config__.serialize_default_excludes)  # type: ignore
        kwargs.setdefault("skip_defaults", self.__config__.serialize_skip_defaults)  # type: ignore
        if self.__config__.force_skip_defaults:  # type: ignore
            kwargs["skip_defaults"] = True

        data = super().dict(**kwargs)

        if encoding is None:
            return data
        elif encoding == "json":
            return json.loads(serialize(data, encoding="json"))
        else:
            raise KeyError(f"Unknown encoding type '{encoding}', valid encoding types: 'json'.")

    def serialize(self,
                  encoding: str,
                  *,
                  include: Optional[Set[str]] = None,
                  exclude: Optional[Set[str]] = None,
                  skip_defaults: bool = False) -> Union[bytes, str]:
        """Generates a serialized representation of the model

        Parameters
        ----------
        encoding : str
            The serialization type, available types are: {'json', 'json-ext', 'msgpack-ext'}
        include : Optional[Set[str]], optional
            Fields to be included in the serialization.
        exclude : Optional[Set[str]], optional
            Fields to be excluded in the serialization.
        skip_defaults : bool, optional
            If True, skips fields that have default values provided.

        Returns
        -------
        Union[bytes, str]
            The serialized model.
        """
        data = self.dict(include=include, exclude=exclude, skip_defaults=skip_defaults)

        return serialize(data, encoding=encoding)

    def compare(self, other: Union['ProtoModel', BaseModel], **kwargs) -> bool:
        """Compares the current object to the provided object recursively.

        Parameters
        ----------
        other : Model
            The model to compare to.
        **kwargs
            Additional kwargs to pass to ``qcelemental.compare_recursive``.

        Returns
        -------
        bool
            True if the objects match.
        """
        return compare_recursive(self, other, **kwargs)

    def __str__(self) -> str:  # lgtm: [py/inheritance/incorrect-overridden-signature]
        if self.__config__.canonical_repr:  # type: ignore
            return super().to_string()
        else:
            return f"{self.__class__.__name__}(ProtoModel)"


class AutodocBaseSettings(BaseSettings, metaclass=PydanticAutodocMeta):
    pass



import re
from functools import lru_cache
from typing import List, Tuple

from ..exceptions import NotAnElementError, ValidationError
from ..periodic_table import periodictable
from .regex import NUCLEUS

_nucleus = re.compile(r'\A' + NUCLEUS + r'\Z', re.IGNORECASE | re.VERBOSE)


@lru_cache(maxsize=512)
def reconcile_nucleus(A: int = None,
                      Z: int = None,
                      E: str = None,
                      mass: float = None,
                      real: bool = None,
                      label: str = None,
                      speclabel: bool = True,
                      nonphysical: bool = False,
                      mtol: float = 1.e-3,
                      verbose: int = 1) -> Tuple[int, int, str, float, bool, str]:
    """Forms consistent set of nucleus descriptors from all information
    from arguments, supplemented by the periodic table. At the least,
    must provide element identity somehow. Defaults to most-abundant
    isotope.

    Parameters
    ----------
    A : int, optional
        Mass number, number of protons and neutrons.
    Z : int, optional
        Atomic number, number of protons.
    E : str, optional
        Element symbol from periodic table.
    mass : float, optional
        Atomic mass [u].
    real : bool, optional
        Whether real or ghost/absent.
    label : str, optional
        Atom label according to :py:data:`qcelemental.molparse.regex.NUCLEUS`.
    speclabel : bool, optional
        If `True`, interpret `label` as potentially full nucleus spec including
        ghosting, isotope, mass, tagging information, e.g., ``@13C_mine`` or
        ``He4@4.01``. If `False`, interpret `label` as only the user/tagging
        extension to nucleus label, e.g. ``_mine`` or ``4`` in the previous examples.
    nonphysical : bool, optional
        When `True`, turns off sanity checking that prevents periodic table
        violations (e.g, light uranium: ``1U@1.007``).
    mtol : float, optional
        How different `mass` can be from a known nuclide mass and still
        merit the mass number assignment. Note that for elements dominated
        by a single isotope, the default may not be tight enough to
        prevent standard atomic weight (abundance-average of isotopes)
        from being labeled as the dominant isotope for A.
    verbose : int, optional
        Quantity of printing.

    Returns
    -------
    A, Z, E, mass, real, userlabel : int, int, str, float, bool, str
        mass number, unless clues don't point to a known nuclide, in which case `-1`.
        atomic number.
        element symbol, capitalized.
        mass value [u].
        real/ghost.
        user portion of `label` if present, else ''.

    Raises
    ------
    qcelemental.NotAnElementError
    qcelemental.ValidationError

    Examples
    --------
    >>> reconcile_nucleus(E='co')
    >>> reconcile_nucleus(Z=27)
    >>> reconcile_nucleus(A=59, Z=27)
    >>> reconcile_nucleus(E='cO', mass=58.933195048)
    >>> reconcile_nucleus(A=59, Z=27, E='CO')
    >>> reconcile_nucleus(A=59, E='cO', mass=58.933195048)
    >>> reconcile_nucleus(label='co')
    >>> reconcile_nucleus(label='59co')
    >>> reconcile_nucleus(label='co@58.933195048')
    >>> reconcile_nucleus(A=59, Z=27, E='cO', mass=58.933195048, label='co@58.933195048')
    >>> reconcile_nucleus(A=59, Z=27, E='cO', mass=58.933195048, label='27@58.933195048')
    >>> reconcile_nucleus(label='27')
    59, 27, 'Co', 58.933195048, True, ''

    >>> reconcile_nucleus(label='co_miNe')
    >>> reconcile_nucleus(label='co_mIne@58.933195048')
    59, 27, 'Co', 58.933195048, True, '_mine'

    >>> reconcile_nucleus(E='cO', mass=58.933)
    >>> reconcile_nucleus(label='cO@58.933')
    59, 27, 'Co', 58.933, True, ''
    >>> assert 59, 27, 'Co', 58.933, True, '' == reconcile_nucleus(E='cO', mass=58.933, mtol=1.e-4))
    AssertionError

    >>> reconcile_nucleus(E='Co', A=60)
    >>> reconcile_nucleus(Z=27, A=60, real=True)
    >>> reconcile_nucleus(E='Co', A=60)
    >>> reconcile_nucleus(Z=27, mass=59.933817059)
    >>> reconcile_nucleus(A=60, Z=27, mass=59.933817059)
    >>> reconcile_nucleus(label='60Co')
    >>> reconcile_nucleus(label='27', mass=59.933817059)
    >>> reconcile_nucleus(label='Co', mass=59.933817059)
    >>> reconcile_nucleus(A=60, label='Co')
    60, 27, 'Co', 59.933817059, True, ''

    >>> reconcile_nucleus(E='Co', A=60, real=False))
    >>> reconcile_nucleus(A=60, Z=27, mass=59.933817059, real=0))
    >>> reconcile_nucleus(label='@60Co'))
    >>> reconcile_nucleus(label='Gh(27)', mass=59.933817059))
    >>> reconcile_nucleus(label='@Co', mass=59.933817059))
    >>> reconcile_nucleus(A=60, label='Gh(Co)'))
    60, 27, 'Co', 59.933817059, False, ''

    >>> reconcile_nucleus(Z=27, mass=200, nonphysical=True)
    60, 27, 'Co', 200.00000000, True, ''

    >>> reconcile_nucleus(mass=60.6, Z=27)
    >>> reconcile_nucleus(mass=60.6, E='Co')
    >>> reconcile_nucleus(mass=60.6, label='27')
    >>> reconcile_nucleus(label='Co@60.6')
    -1, 27, 'Co', 60.6, True, ''

    >>> reconcile_nucleus(mass=60.6, Z=27, A=61))
    >>> reconcile_nucleus(A=80, Z=27)
    >>> reconcile_nucleus(Z=27, mass=200)
    >>> reconcile_nucleus(Z=27, mass=-200, nonphysical=True)
    >>> reconcile_nucleus(Z=-27, mass=200, nonphysical=True)
    >>> reconcile_nucleus(Z=1, label='he')
    >>> reconcile_nucleus(A=4, label='3he')
    >>> reconcile_nucleus(label='@U', real=True)
    >>> reconcile_nucleus(label='U', real=False)
    ValidationError

    """

    # <<< define functions
    log_text = verbose >= 2

    def reconcile(exact, tests, feature):
        """Returns a member from `exact` that passes all `tests`, else raises error for `feature`."""

        for candidate in exact:
            assessment = [fn(candidate) for fn in tests]
            if log_text:
                text.append("""Assess {} candidate {}: {} --> {}""".format(feature, candidate, assessment,
                                                                           all(assessment)))
            if all(assessment):
                return candidate

        err = """Inconsistent or unspecified {}: A: {}, Z: {}, E: {}, mass: {}, real: {}, label: {}""".format(
            feature, A, Z, E, mass, real, label)
        if verbose > -1:
            print('\n\n' + '\n'.join(text))
        raise ValidationError(err)

    def offer_element_symbol(e):
        """Given an element, what can be suggested and asserted about Z, A, mass?"""

        _Z = periodictable.to_Z(e)
        offer_atomic_number(_Z)

    def offer_atomic_number(z):
        """Given an atomic number, what can be suggested and asserted about Z, A, mass?"""

        z = int(z)

        z_symbol = periodictable.to_E(z)
        z_mass = periodictable.to_mass(z)
        z_a = periodictable.to_A(z)

        z_a2mass = periodictable._el2a2mass[z_symbol]
        z_a2mass_min = min(z_a2mass.keys())
        z_a2mass_max = max(z_a2mass.keys())
        z_mass2a_min = min(z_a2mass.values())
        z_mass2a_max = max(z_a2mass.values())

        Z_exact.append(z)
        Z_range.append(lambda x, z=z: x == z)

        A_exact.append(z_a)
        if nonphysical:
            A_range.append(lambda x: x == -1 or x >= 1)
            if log_text:
                text.append("""For A, input Z: {}, requires 1 < A or -1, nonphysical""".format(z))
        else:
            A_range.append(lambda x, amin=z_a2mass_min, amax=z_a2mass_max: x == -1 or (x >= amin and x <= amax))
            if log_text:
                text.append(
                    """For A, input Z: {} requires {} < A < {} or -1, the known mass number range for element""".
                    format(z, z_a2mass_min, z_a2mass_max))

        m_exact.append(z_mass)
        if nonphysical:
            m_range.append(lambda x: x > 0.5)
            if log_text:
                text.append("""For mass, input Z: {}, requires 0.5 < mass, nonphysical""".format(z))
        else:
            m_range.append(lambda x, mmin=z_mass2a_min, mmax=z_mass2a_max: x >= mmin - mmtol and x <= mmax + mmtol)
            if log_text:
                text.append(
                    """For mass, input Z: {} requires {} < mass < {} +/-{}, the known mass range for element""".format(
                        z, z_mass2a_min, z_mass2a_max, mmtol))

    def offer_mass_number(z, a):
        """Given a mass number and element, what can be suggested and asserted about A, mass?"""

        a = int(a)
        a_eliso = periodictable.to_E(z) + str(a)
        a_mass = periodictable.to_mass(a_eliso)

        A_exact.append(a)
        A_range.append(lambda x, a=a: x == a)
        if log_text:
            text.append("""For A, inputs Z: {}, A: {} require A == {}.""".format(z, a, a))

        m_exact.append(a_mass)
        m_range.append(lambda x, a_mass=a_mass: abs(x - a_mass) < mtol)
        if log_text:
            text.append("""For mass, inputs Z: {}, A: {} require abs(mass - {}) < {}""".format(z, a, a_mass, mtol))

    def offer_mass_value(z, m):
        """Given a mass and element, what can be suggested and asserted about A, mass?"""

        m = float(m)
        m_a = int(round(m, 0))
        m_eliso = periodictable.to_E(z) + str(m_a)

        try:
            if abs(periodictable.to_mass(m_eliso) - m) > mtol:
                # only offer A if known nuclide. C@12.4 != 12C
                m_a = -1
        except NotAnElementError:
            m_a = -1

        A_exact.append(m_a)
        A_range.append(lambda x, m_a=m_a: x == m_a)
        if log_text:
            text.append("""For A, inputs Z: {}, m: {} require A == {}""".format(z, m, m_a))

        m_exact.append(m)
        m_range.append(lambda x, m=m: x == m)
        if log_text:
            text.append("""For mass, inputs Z: {}, m: {} require m == {}""".format(z, m, m))

    def offer_reality(rgh):
        r_exact.append(rgh)
        r_range.append(lambda x, rgh=rgh: x == rgh)
        if log_text:
            text.append("""For real/ghost, input rgh: {} requires rgh == {}""".format(rgh, rgh))

    def offer_user_label(lbl):
        lbl = str(lbl).lower()
        l_exact.append(lbl)
        l_range.append(lambda x, lbl=lbl: x == lbl)
        if log_text:
            text.append("""For user label, input lbl: {} requires lbl == {}""".format(lbl, lbl))

    # <<< initialize

    text = ['', """--> Inp: A={}, Z={}, E={}, mass={}, real={}, label={}""".format(A, Z, E, mass, real, label)]

    Z_exact: List = []  # *_exact are candidates for the final value
    Z_range: List = []  # *_range are tests that the final value must pass to be valid
    A_exact: List = []
    A_range: List = []
    m_exact: List = []
    m_range: List = []

    r_exact = [True]  # default real/ghost is real
    r_range: List = []
    l_exact = ['']  # default user label is empty string
    l_range: List = []
    mmtol = 0.5  # tolerance for mass outside known masses for element

    # <<< collect evidence for Z/A/m, then reconcile Z

    if Z is not None:
        offer_atomic_number(Z)

    if E is not None:
        offer_element_symbol(E)

    if label is not None and speclabel is True:
        lbl_A, lbl_Z, lbl_E, lbl_mass, lbl_real, lbl_user = parse_nucleus_label(label)

        if lbl_Z is not None:
            offer_atomic_number(lbl_Z)
        if lbl_E is not None:
            offer_element_symbol(lbl_E)

    Z_final = reconcile(Z_exact, Z_range, 'atomic number')
    E_final = periodictable.to_E(Z_final)

    # <<< collect more evidence for A/m, then reconcile them

    if A is not None:
        offer_mass_number(Z_final, A)

    if mass is not None:
        offer_mass_value(Z_final, mass)

    if real is not None:
        offer_reality(real)

    if label is not None:
        if speclabel:
            offer_reality(lbl_real)
            if lbl_A is not None:
                offer_mass_number(Z_final, lbl_A)
            if lbl_mass is not None:
                offer_mass_value(Z_final, lbl_mass)
            if lbl_user is not None:
                offer_user_label(lbl_user)
        else:
            offer_user_label(label)

    mass_final = reconcile(m_exact, m_range, 'mass')
    A_final = reconcile(A_exact, A_range, 'mass number')
    real_final = reconcile(r_exact, r_range, 'real/ghost')
    user_final = reconcile(l_exact, l_range, 'user label')

    if log_text:
        text.append("""<-- Out: A={}, Z={}, E={}, mass={}, real={}, user={}""".format(
            A_final, Z_final, E_final, mass_final, real_final, user_final))

    if log_text:
        print('\n'.join(text))

    return (A_final, Z_final, E_final, mass_final, real_final, user_final)


def parse_nucleus_label(label):
    """Separate molecule nucleus string into fields.

    Parameters
    ----------
    label : str
        Conveys at least element and ghostedness and possibly isotope, mass, and
        user info in accordance with :py:data:`qcelemental.molparse.regex.NUCLEUS`.

    Returns
    -------
    A, Z, E, mass, real, user : int or None, int or None, str or None, float or None, bool, str or None
        Field breakdown of `label`.

    Raises
    ------
    qcelemental.ValidationError
        If `label` does not match NUCLEUS.

    Examples
    --------
    >>> parse_nucleus_label('@ca_miNe')
    None, None, 'ca', None False, '_miNe'

    >>> parse_nucleus_label('Gh(Ca_mine)')
    None, None, 'Ca', None '_mine', False

    >>> parse_nucleus_label('@Ca_mine@1.07')
    None, None, 'Ca', 1.07 False, '_mine'

    >>> parse_nucleus_label('Gh(cA_MINE@1.07)')
    None, None, 'cA', 1.07 False, '_MINE'

    >>> parse_nucleus_label('@40Ca_mine@1.07')
    40, None, 'Ca', 1.07 False, '_mine'

    >>> parse_nucleus_label('Gh(40Ca_mine@1.07)')
    40, None, 'Ca', 1.07 False, '_mine'

    >>> parse_nucleus_label('444lu333@4.0')
    444, None, 'lu', 4.0 True, '333'

    >>> parse_nucleus_label('@444lu333@4.4')
    444, None, 'lu', 4.4 False, '333'

    >>> parse_nucleus_label('8i')
    8, None, 'i', None True, None

    >>> parse_nucleus_label('53_mI4')
    None, 53, None, None True, '_mI4'

    >>> parse_nucleus_label('@5_MINEs3@4.4')
    None, 5, None, 4.4 False, '_MINEs3'

    >>> parse_nucleus_label('Gh(555_mines3@0.1)')
    None, 555, None, 0.1 False, '_mines3'

    """
    matchobj = _nucleus.match(label)

    if matchobj:
        real = not (matchobj.group('gh1') or matchobj.group('gh2'))

        if matchobj.group('A'):
            A = int(matchobj.group('A'))
        else:
            A = None

        if matchobj.group('Z'):
            Z = int(matchobj.group('Z'))
        else:
            Z = None

        E = matchobj.group('E')

        if matchobj.group('user1'):
            user = matchobj.group('user1')
        elif matchobj.group('user2'):
            user = matchobj.group('user2')
        else:
            user = None

        if matchobj.group('mass'):
            mass = float(matchobj.group('mass'))
        else:
            mass = None
    else:
        raise ValidationError("""Nucleus label is not parseable: {}""".format(label))

    return A, Z, E, mass, real, user



from .autodocs import auto_gen_docs_on_demand, get_base_docs
from .gph_uno_bipartite import uno
from .importing import parse_version, safe_version, which, which_import
from .internal import provenance_stamp
from .itertools import unique_everseen
#from .mpl import plot_coord
from .misc import (compute_angle, compute_dihedral, compute_distance, distance_matrix, filter_comments,
                   measure_coordinates, standardize_efp_angles_units, unnp, update_with_error)
from .np_blockwise import blockwise_contract, blockwise_expand
from .np_rand3drot import random_rotation_matrix
from .scipy_hungarian import linear_sum_assignment
from .serialization import (deserialize, json_dumps, json_loads, jsonext_dumps, jsonext_loads, msgpackext_dumps,
                            msgpackext_loads, serialize)



from decimal import Decimal

import pytest

import qcelemental


# 2014 CODATA values
@pytest.mark.parametrize("from_unit, to_unit, expected", [

    # First check the 7 canonical NIST conversions
    ("hartree", "Hz", "6.579683920711e15"),
    ("amu", "hartree", "3.4231776902e7"),
    ("hartree", "kilogram", "4.850870129e-35"),
    ("hartree", "J/mol", "2625499.638"),
    ("hartree", "eV", "27.21138602"),
    ("hartree", "wavenumber", "2.194746313702e5"),
    ("hartree", "kelvin", "3.1577513e5"),

    # Check other values
    ("millihartree", "Hz", "6.579683920711e9"),
    ("mhartree", "Hz", "6.579683920711e9"),
    ("hartree", "microgram", "4.850870129e-26"),
    ("microhartree", "wavenumber", "2.194746313702e-7"),
    ("uhartree", "wavenumber", "2.194746313702e-7"),
    ("kilohartree", "kiloeV", "27.21138602"),
    ("hartree", "kiloeV", "0.02721138602"),
    ("millihartree", "J/mol", "2625.499638"),
    ("hartree", "kJ/mol", "2625.499638"),
    ("hartree", "J", "4.359744650e-18"),
    ("1/m", "hartree", "4.556335e-8"),
]) # yapf: disable
def test_unit_conversion_nist(from_unit, to_unit, expected):

    # Build values and quantize to the tolerance
    expected = Decimal(expected)
    from_to_value = Decimal(qcelemental.constants.conversion_factor(from_unit, to_unit)).quantize(expected)
    assert from_to_value == expected

@pytest.mark.parametrize("from_unit, to_unit, expected", [
    ("hartree", "kcal/mol", "627.509474"),
    ("hartree", "kJ/mol", "2625.499638"),
    ("bohr", "angstrom", "0.52917721067"),
    ("bohr", "nanometer", "5.2917721067e-2"),
    ("bohr", "cm", "5.2917721067e-9"),
    ("bohr", "miles", "3.288154743E-14"),
    ("amu", "kg", "1.660539040e-27"),
    ("amu", "g", "1.660539040e-24"),
    ("Bohr", "bohr", "1.0"),
    ("Angstrom", "Angstrom", "1.0"),
]) # yapf: disable
def test_unit_conversion_other(from_unit, to_unit, expected):

    # Build values and quantize to the tolerance
    expected = Decimal(expected)
    from_to_value = Decimal(qcelemental.constants.conversion_factor(from_unit, to_unit)).quantize(expected)
    assert from_to_value == expected

    # Build inverse
    inv_expected = 1 / Decimal(expected)
    to_from_value = Decimal(qcelemental.constants.conversion_factor(to_unit, from_unit))

    # Expected to a relative tolerance as the number of digits plus two for rounding
    # Using float comparisons as we are taking an (1 / float) inverse in the conversion code
    rel_tol = float("10e-{}".format(len(expected.as_tuple().digits) + 2))
    assert pytest.approx(float(inv_expected), rel_tol) == float(to_from_value)


def test_quantities_smoke():
    """
    Smoke test to ensure Quantities are correctly returned
    """
    assert 5 == qcelemental.constants.Quantity("5 kcal").magnitude



import os
import shutil
import sys
from typing import Union


def which_import(module: str, *, return_bool: bool = False, raise_error: bool = False,
                 raise_msg: str = None) -> Union[bool, None, str]:
    """Tests to see if a Python module is available.

    Returns
    -------
    str or None
        By default, returns `__init__.py`-like path if module found or `None` if not.
    bool
        When `return_bool=True`, returns whether or not found.

    Raises
    ------
    ModuleNotFoundError
        When `raises_error=True` and module not found. Raises generic message plus any `raise_msg`.

    """
    import importlib
    module_spec = importlib.util.find_spec(module)

    if module_spec is None:
        if raise_error:
            raise ModuleNotFoundError(
                f"Python module '{module}' not found in envvar PYTHONPATH.{' ' + raise_msg if raise_msg else ''}")
        elif return_bool:
            return False
        else:
            return None
    else:
        if return_bool:
            return True
        else:
            return module_spec.origin


def which(command: str,
          *,
          return_bool: bool = False,
          raise_error: bool = False,
          raise_msg: str = None,
          env: str = None) -> Union[bool, None, str]:
    """Test to see if a command is available.

    Returns
    -------
    str or None
        By default, returns command path if command found or `None` if not.
        Environment is $PATH or `os.pathsep`-separated `env`, less any None values.
    bool
        When `return_bool=True`, returns whether or not found.

    Raises
    ------
    ModuleNotFoundError
        When `raises_error=True` and command not found. Raises generic message plus any `raise_msg`.

    """
    if env is None:
        lenv = {'PATH': os.pathsep + os.environ.get('PATH', '') + os.path.dirname(sys.executable)}
    else:
        lenv = {'PATH': os.pathsep.join([os.path.abspath(x) for x in env.split(os.pathsep) if x != ''])}
    lenv = {k: v for k, v in lenv.items() if v is not None}

    ans = shutil.which(command, mode=os.F_OK | os.X_OK, path=lenv['PATH'])

    if raise_error and ans is None:
        raise ModuleNotFoundError(
            f"Command '{command}' not found in envvar PATH.{' ' + raise_msg if raise_msg else ''}")

    if return_bool:
        return bool(ans)
    else:
        return ans


def safe_version(*args, **kwargs) -> str:
    """
    Package resources is a very slow load
    """
    import pkg_resources
    return pkg_resources.safe_version(*args, **kwargs)


def parse_version(*args, **kwargs):
    """
    Package resources is a very slow load
    """
    import pkg_resources
    return pkg_resources.parse_version(*args, **kwargs)



"""
Molecule Object Model
"""

import collections
import hashlib
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from pydantic import Schema, constr, validator

from ..molparse import from_arrays, from_schema, from_string, to_schema, to_string
from ..periodic_table import periodictable
from ..physical_constants import constants
from ..testing import compare, compare_values
from ..util import deserialize, measure_coordinates, provenance_stamp, which_import, msgpackext_loads
from .basemodels import ProtoModel
from .common_models import Provenance, qcschema_molecule_default
from .types import Array

# Rounding quantities for hashing
GEOMETRY_NOISE = 8
MASS_NOISE = 6
CHARGE_NOISE = 4

_extension_map = {
    ".npy": "numpy",
    ".json": "json",
    ".xyz": "xyz",
    ".psimol": "psi4",
    ".psi4": "psi4",
    ".msgpack": "msgpack",
}


def float_prep(array, around):
    """
    Rounds floats to a common value and build positive zeros to prevent hash conflicts.
    """
    if isinstance(array, (list, np.ndarray)):
        # Round array
        array = np.around(array, around)
        # Flip zeros
        array[np.abs(array) < 5**(-(around + 1))] = 0

    elif isinstance(array, (float, int)):
        array = round(array, around)
        if array == -0.0:
            array = 0.0
    else:
        raise TypeError("Type '{}' not recognized".format(type(array).__name__))

    return array


class Identifiers(ProtoModel):
    """Canonical chemical identifiers"""

    molecule_hash: Optional[str] = None
    molecular_formula: Optional[str] = None
    smiles: Optional[str] = None
    inchi: Optional[str] = None
    inchikey: Optional[str] = None
    canonical_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_mapped_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_smiles: Optional[str] = None
    canonical_smiles: Optional[str] = None

    class Config(ProtoModel.Config):
        serialize_skip_defaults = True


class Molecule(ProtoModel):
    """
    A QCSchema representation of a Molecule. This model contains
    data for symbols, geometry, connectivity, charges, fragmentation, etc while also supporting a wide array of I/O and manipulation capabilities.

    Molecule objects geometry, masses, and charges are truncated to 8, 6, and 4 decimal places respectively to assist with duplicate detection.
    """

    schema_name: constr(strip_whitespace=True, regex=qcschema_molecule_default) = Schema(  # type: ignore
        qcschema_molecule_default,
        description=(f"The QCSchema specification this model conforms to. Explicitly fixed as "
                     f"{qcschema_molecule_default}."))
    schema_version: int = Schema(  # type: ignore
        2, description="The version number of ``schema_name`` that this Molecule model conforms to.")
    validated: bool = Schema(  # type: ignore
        False,
        description="A boolean indicator (for speed purposes) that the input Molecule data has been previously checked "
        "for schema (data layout and type) and physics (e.g., non-overlapping atoms, feasible "
        "multiplicity) compliance. This should be False in most cases. A ``True`` setting "
        "should only ever be set by the constructor for this class itself or other trusted sources such as "
        "a Fractal Server or previously serialized Molecules.")

    # Required data
    symbols: Array[str] = Schema(  # type: ignore
        ...,
        description="An ordered (nat,) array-like object of atomic elemental symbols of shape (nat,). The index of "
        "this attribute sets atomic order for all other per-atom setting like ``real`` and the first "
        "dimension of ``geometry``. Ghost/Virtual atoms must have an entry in this array-like and are "
        "indicated by the matching the 0-indexed indices in ``real`` field.")
    geometry: Array[float] = Schema(  # type: ignore
        ...,
        description="An ordered (nat,3) array-like for XYZ atomic coordinates [a0]. "
        "Atom ordering is fixed; that is, a consumer who shuffles atoms must not reattach the input "
        "(pre-shuffling) molecule schema instance to any output (post-shuffling) per-atom results "
        "(e.g., gradient). Index of the first dimension matches the 0-indexed indices of all other "
        "per-atom settings like ``symbols`` and ``real``."
        "\n"
        "Can also accept array-likes which can be mapped to (nat,3) such as a 1-D list of length 3*nat, "
        "or the serialized version of the array in (3*nat,) shape; all forms will be reshaped to "
        "(nat,3) for this attribute.")

    # Molecule data
    name: Optional[str] = Schema(  # type: ignore
        None,
        description="A common or human-readable name to assign to this molecule. Can be arbitrary.")
    identifiers: Optional[Identifiers] = Schema(  # type: ignore
        None,
        description="An optional dictionary of additional identifiers by which this Molecule can be referenced, "
        "such as INCHI, canonical SMILES, etc. See the :class:``Identifiers`` model for more details.")
    comment: Optional[str] = Schema(  # type: ignore
        None,
        description="Additional comments for this Molecule. Intended for pure human/user consumption "
        "and clarity.")
    molecular_charge: float = Schema(0.0, description="The net electrostatic charge of this Molecule.")  # type: ignore
    molecular_multiplicity: int = Schema(1, description="The total multiplicity of this Molecule.")  # type: ignore

    # Atom data
    masses: Optional[Array[float]] = Schema(  # type: ignore
        None,
        description="An ordered 1-D array-like object of atomic masses [u] of shape (nat,). Index order "
        "matches the 0-indexed indices of all other per-atom settings like ``symbols`` and ``real``. If "
        "this is not provided, the mass of each atom is inferred from their most common isotope. If this "
        "is provided, it must be the same length as ``symbols`` but can accept ``None`` entries for "
        "standard masses to infer from the same index in the ``symbols`` field.")
    real: Optional[Array[bool]] = Schema(  # type: ignore
        None,
        description="An ordered 1-D array-like object of shape (nat,) indicating if each atom is real (``True``) or "
        "ghost/virtual (``False``). Index "
        "matches the 0-indexed indices of all other per-atom settings like ``symbols`` and the first "
        "dimension of ``geometry``. If this is not provided, all atoms are assumed to be real (``True``)."
        "If this is provided, the reality or ghostality of every atom must be specified.")
    atom_labels: Optional[Array[str]] = Schema(  # type: ignore
        None,
        description="Additional per-atom labels as a 1-D array-like of of strings of shape (nat,). Typical use is in "
        "model conversions, such as Elemental <-> Molpro and not typically something which should be user "
        "assigned. See the ``comments`` field for general human-consumable text to affix to the Molecule.")
    atomic_numbers: Optional[Array[np.int16]] = Schema(  # type: ignore
        None,
        description="An optional ordered 1-D array-like object of atomic numbers of shape (nat,). Index "
        "matches the 0-indexed indices of all other per-atom settings like ``symbols`` and ``real``. "
        "Values are inferred from the ``symbols`` list if not explicitly set.")
    mass_numbers: Optional[Array[np.int16]] = Schema(  # type: ignore
        None,
        description="An optional ordered 1-D array-like object of atomic *mass* numbers of shape (nat). Index "
        "matches the 0-indexed indices of all other per-atom settings like ``symbols`` and ``real``. "
        "Values are inferred from the most common isotopes of the ``symbols`` list if not explicitly set.")

    # Fragment and connection data
    connectivity: Optional[List[Tuple[int, int, float]]] = Schema(  # type: ignore
        None,
        description="The connectivity information between each atom in the ``symbols`` array. Each entry in this "
        "list is a Tuple of ``(atom_index_A, atom_index_B, bond_order)`` where the ``atom_index`` "
        "matches the 0-indexed indices of all other per-atom settings like ``symbols`` and ``real``.")
    fragments: Optional[List[Array[np.int32]]] = Schema(  # type: ignore
        None,
        description="An indication of which sets of atoms are fragments within the Molecule. This is a list of shape "
        "(nfr) of 1-D array-like objects of arbitrary length. Each entry in the list indicates a new "
        "fragment. The index "
        "of the list matches the 0-indexed indices of ``fragment_charges`` and "
        "``fragment_multiplicities``. The 1-D array-like objects are sets of atom indices indicating the "
        "atoms which compose the fragment. The atom indices match the 0-indexed indices of all other "
        "per-atom settings like ``symbols`` and ``real``.")
    fragment_charges: Optional[List[float]] = Schema(  # type: ignore
        None,
        description="The total charge of each fragment in the ``fragments`` list of shape (nfr,). The index of this "
        "list matches the 0-index indices of ``fragment`` list. Will be filled in based on a set of rules "
        "if not provided (and ``fragments`` are specified).")
    fragment_multiplicities: Optional[List[int]] = Schema(  # type: ignore
        None,
        description="The multiplicity of each fragment in the ``fragments`` list of shape (nfr,). The index of this "
        "list matches the 0-index indices of ``fragment`` list. Will be filled in based on a set of "
        "rules if not provided (and ``fragments`` are specified).")

    # Orientation
    fix_com: bool = Schema(  # type: ignore
        False,
        description="An indicator which prevents pre-processing the Molecule object to translate the Center-of-Mass "
        "to (0,0,0) in euclidean coordinate space. Will result in a different ``geometry`` than the "
        "one provided if False.")
    fix_orientation: bool = Schema(  # type: ignore
        False,
        description="An indicator which prevents pre-processes the Molecule object to orient via the inertia tensor."
        "Will result in a different ``geometry`` than the one provided if False.")
    fix_symmetry: Optional[str] = Schema(  # type: ignore
        None,
        description="Maximal point group symmetry which ``geometry`` should be treated. Lowercase.")
    # Extra
    provenance: Provenance = Schema(  # type: ignore
        provenance_stamp(__name__),
        description="The provenance information about how this Molecule (and its attributes) were generated, "
        "provided, and manipulated.")
    id: Optional[Any] = Schema(  # type: ignore
        None,
        description="A unique identifier for this Molecule object. This field exists primarily for Databases "
        "(e.g. Fractal's Server) to track and lookup this specific object and should virtually "
        "never need to be manually set.")
    extras: Dict[str, Any] = Schema(  # type: ignore
        None, description="Extra information to associate with this Molecule.")

    class Config(ProtoModel.Config):
        serialize_skip_defaults = True

    def __init__(self, orient: bool = False, validate: Optional[bool] = None, **kwargs: Any) -> None:
        """Initializes the molecule object from dictionary-like values.

        Parameters
        ----------
        orient : bool, optional
            If True, orientates the Molecule to a common reference frame.
        validate : Optional[bool], optional
            If ``None`` validation is always applied unless the ``validated`` flag is set. Otherwise uses the boolean to decide to validate the Molecule or not.
        **kwargs : Any
            The values of the Molecule object attributes.
        """
        if validate is None:
            validate = not kwargs.get("validated", False)

        if validate:
            kwargs["schema_name"] = kwargs.pop("schema_name", "qcschema_molecule")
            kwargs["schema_version"] = kwargs.pop("schema_version", 2)
            # original_keys = set(kwargs.keys())  # revive when ready to revisit sparsity

            schema = to_schema(from_schema(kwargs), dtype=kwargs["schema_version"], copy=False, np_out=True)

            kwargs["validated"] = True
            kwargs = {**kwargs, **schema}  # Allow any extra fields
            validate = True

        super().__init__(**kwargs)

        # We are pulling out the values *explicitly* so that the pydantic skip_defaults works as expected
        # All attributes set bellow are equivalent to the default set.
        values = self.__dict__

        natoms = values["geometry"].shape[0]
        if validate:
            values["symbols"] = np.core.defchararray.title(self.symbols)  # Title case for consistency

        if values["masses"] is None:  # Setup masses before fixing the orientation
            values["masses"] = np.array([periodictable.to_mass(x) for x in values["symbols"]])

        if values["real"] is None:
            values["real"] = np.ones(natoms, dtype=bool)

        if orient:
            values["geometry"] = float_prep(self._orient_molecule_internal(), GEOMETRY_NOISE)
        elif validate:
            values["geometry"] = float_prep(values["geometry"], GEOMETRY_NOISE)

        # Cleanup un-initialized variables  (more complex than Pydantic Validators allow)
        if values["fragments"] is None:
            values["fragments"] = [np.arange(natoms, dtype=np.int32)]
            values["fragment_charges"] = [values["molecular_charge"]]
            values["fragment_multiplicities"] = [values["molecular_multiplicity"]]
        else:
            if values["fragment_charges"] is None:
                if np.isclose(values["molecular_charge"], 0.0):
                    values["fragment_charges"] = [0 for _ in values["fragments"]]
                else:
                    raise KeyError("Fragments passed in, but not fragment charges for a charged molecule.")

            if values["fragment_multiplicities"] is None:
                if values["molecular_multiplicity"] == 1:
                    values["fragment_multiplicities"] = [1 for _ in values["fragments"]]
                else:
                    raise KeyError("Fragments passed in, but not fragment multiplicities for a non-singlet molecule.")

    @validator('geometry')
    def _must_be_3n(cls, v, values, **kwargs):
        n = len(values['symbols'])
        try:
            v = v.reshape(n, 3)
        except (ValueError, AttributeError):
            raise ValueError("Geometry must be castable to shape (N,3)!")
        return v

    @validator('masses', 'real', whole=True)
    def _must_be_n(cls, v, values, **kwargs):
        n = len(values['symbols'])
        if len(v) != n:
            raise ValueError("Masses and Real must be same number of entries as Symbols")
        return v

    @validator('real', whole=True)
    def _populate_real(cls, v, values, **kwargs):
        # Can't use geometry here since its already been validated and not in values
        n = len(values['symbols'])
        if len(v) == 0:
            v = np.array([True for _ in range(n)])
        return v

    @validator('fragment_charges', 'fragment_multiplicities', whole=True)
    def _must_be_n_frag(cls, v, values, **kwargs):
        if 'fragments' in values:
            n = len(values['fragments'])
            if len(v) != n:
                raise ValueError("Fragment Charges and Fragment Multiplicities"
                                 " must be same number of entries as Fragments")
        else:
            raise ValueError("Cannot have Fragment Charges or Fragment Multiplicities " "without Fragments")
        return v

    @validator('connectivity')
    def _min_zero(cls, v):
        if v < 0:
            raise ValueError("Connectivity entries must be greater than 0")
        return v

    @property
    def hash_fields(self):
        return [
            "symbols", "masses", "molecular_charge", "molecular_multiplicity", "real", "geometry", "fragments",
            "fragment_charges", "fragment_multiplicities", "connectivity"
        ]

### Non-Pydantic API functions

    def show(self, *, style: Union[str, Dict[str, Any]] = "ball_and_stick",
             canvas: Tuple[int, int] = (400, 400)) -> 'py3Dmol.view':  # type: ignore
        """Creates a 3D representation of a moleucle that can be manipulated in Jupyter Notebooks and exported as
        images (`.png`).

        Parameters
        ----------
        style : Union[str, Dict[str, Any]], optional
            Either 'ball_and_stick' or 'stick' style representations or a valid py3Dmol style dictionary.
        canvas : Tuple[int, int], optional
            The (width, height) of the display canvas in pixels

        Returns
        -------
        py3Dmol.view
            A py3dMol view object of the molecule

        """
        if not which_import("py3Dmol", return_bool=True):
            raise ModuleNotFoundError(
                f"Python module py3DMol not found. Solve by installing it: `conda install -c conda-forge py3dmol`"
            )  # pragma: no cover

        import py3Dmol  # type: ignore

        if isinstance(style, dict):
            pass
        elif style == 'ball_and_stick':
            style = {'stick': {'radius': 0.2}, 'sphere': {'scale': 0.3}}
        elif style == 'stick':
            style = {'stick': {}}
        else:
            raise KeyError(f"Style '{style}' not recognized.")

        xyzview = py3Dmol.view(width=canvas[0], height=canvas[1])

        xyzview.addModel(self.to_string("xyz", units="angstrom"), 'xyz')
        xyzview.setStyle(style)
        xyzview.zoomTo()
        return xyzview

    def measure(self, measurements: Union[List[int], List[List[int]]], *,
                degrees: bool = True) -> Union[float, List[float]]:
        """
        Takes a measurement of the moleucle from the indicies provided.

        Parameters
        ----------
        measurements : Union[List[int], List[List[int]]]
            Either a single list of indices or multiple. Return a distance, angle, or dihedral depending if
            2, 3, or 4 indices is provided, respectively. Values are returned in Bohr (distance) or degree.
        degrees : bool, optional
            Returns degrees by default, radians otherwise.

        Returns
        -------
        Union[float, List[float]]
            Either a value or list of the measured values.
        """

        return measure_coordinates(self.geometry, measurements, degrees=degrees)

    def orient_molecule(self):
        """
        Centers the molecule and orients via inertia tensor before returning a new Molecule
        """
        return Molecule(orient=True, **self.dict())

    def compare(self, other):
        """
        Checks if two molecules are identical. This is a molecular identity defined
        by scientific terms, and not programing terms, so it's less rigorous than
        a programmatic equality or a memory equivalent `is`.
        """

        if isinstance(other, dict):
            other = Molecule(orient=False, **other)
        elif isinstance(other, Molecule):
            pass
        else:
            raise TypeError("Comparison molecule not understood of type '{}'.".format(type(other)))

        match = True
        match &= np.array_equal(self.symbols, other.symbols)
        match &= np.allclose(self.masses, other.masses, atol=(10**-MASS_NOISE))
        match &= np.array_equal(self.real, other.real)
        match &= np.array_equal(self.fragments, other.fragments)
        match &= np.allclose(self.fragment_charges, other.fragment_charges, atol=(10**-CHARGE_NOISE))
        match &= np.array_equal(self.fragment_multiplicities, other.fragment_multiplicities)

        match &= np.allclose(self.molecular_charge, other.molecular_charge, atol=(10**-CHARGE_NOISE))
        match &= np.array_equal(self.molecular_multiplicity, other.molecular_multiplicity)
        match &= np.allclose(self.geometry, other.geometry, atol=(10**-GEOMETRY_NOISE))
        return match

    def pretty_print(self):
        """Print the molecule in Angstroms. Same as :py:func:`print_out` only always in Angstroms.
        (method name in libmints is print_in_angstrom)

        """
        text = ""

        text += """    Geometry (in {0:s}), charge = {1:.1f}, multiplicity = {2:d}:\n\n""".format(
            'Angstrom', self.molecular_charge, self.molecular_multiplicity)
        text += """       Center              X                  Y                   Z       \n"""
        text += """    ------------   -----------------  -----------------  -----------------\n"""

        for i in range(len(self.geometry)):
            text += """    {0:8s}{1:4s} """.format(self.symbols[i], "" if self.real[i] else "(Gh)")
            for j in range(3):
                text += """  {0:17.12f}""".format(self.geometry[i][j] *
                                                  constants.conversion_factor("bohr", "angstroms"))
            text += "\n"
        # text += "\n"

        return text

    def get_fragment(self,
                     real: Union[int, List],
                     ghost: Optional[Union[int, List]] = None,
                     orient: bool = False,
                     group_fragments: bool = True) -> 'Molecule':
        """Get new Molecule with fragments preserved, dropped, or ghosted.

        Parameters
        ----------
        real
            Fragment index or list of indices (0-indexed) to be real atoms in new Molecule.
        ghost
            Fragment index or list of indices (0-indexed) to be ghost atoms (basis fns only) in new Molecule.
        orient
            Whether or not to align (inertial frame) and phase geometry upon new Molecule instantiation
            (according to _orient_molecule_internal)?
        group_fragments
            Whether or not to group real fragments at the start of the atom list and ghost fragments toward the back.
            Previous to ``v0.5``, this was always effectively True. True is handy for finding duplicate
            (atom-order-independent) molecules by hash. False preserves fragment order (though collapsing gaps for
            absent fragments) like Psi4's ``extract_subsets``. False is handy for gradients where atom order of
            returned values matters.

        Returns
        -------
        mol
            New ``py::class:qcelemental.model.Molecule`` with ``self``\'s fragments present, ghosted, or absent.

        """
        if isinstance(real, int):
            real = [real]

        if isinstance(ghost, int):
            ghost = [ghost]
        elif ghost is None:
            ghost = []

        constructor_dict: Dict = {}

        ret_name = (self.name if self.name is not None else '') + " (" + str(real) + "," + str(ghost) + ")"
        constructor_dict["name"] = ret_name
        # ret = Molecule(None, name=ret_name)

        if len(set(real) & set(ghost)):
            raise TypeError("Molecule:get_fragment: real and ghost sets are overlapping! ({0}, {1}).".format(
                str(real), str(ghost)))

        geom_blocks = []
        symbols = []
        masses = []
        real_atoms = []
        fragments = []
        fragment_charges = []
        fragment_multiplicities = []
        atom_size = 0

        if group_fragments:

            # Loop through the real blocks
            frag_start = 0
            for frag in real:
                frag_size = len(self.fragments[frag])
                geom_blocks.append(self.geometry[self.fragments[frag]])

                for idx in self.fragments[frag]:
                    symbols.append(self.symbols[idx])
                    real_atoms.append(True)
                    masses.append(self.masses[idx])

                fragments.append(list(range(frag_start, frag_start + frag_size)))
                frag_start += frag_size

                fragment_charges.append(float(self.fragment_charges[frag]))
                fragment_multiplicities.append(self.fragment_multiplicities[frag])

            # Set charge and multiplicity
            constructor_dict["molecular_charge"] = sum(fragment_charges)
            constructor_dict["molecular_multiplicity"] = sum(x - 1 for x in fragment_multiplicities) + 1

            # Loop through the ghost blocks
            for frag in ghost:
                frag_size = len(self.fragments[frag])
                geom_blocks.append(self.geometry[self.fragments[frag]])

                for idx in self.fragments[frag]:
                    symbols.append(self.symbols[idx])
                    real_atoms.append(False)
                    masses.append(self.masses[idx])

                fragments.append(list(range(frag_start, frag_start + frag_size)))
                frag_start += frag_size

                fragment_charges.append(0)
                fragment_multiplicities.append(1)

        else:
            # List[Array[np.int32]]
            at2fr: List[Union[int, None]] = [None] * len(self.symbols)
            for ifr, fr in enumerate(self.fragments):
                for iat in fr:
                    at2fr[iat] = ifr

            at2at: List[Union[int, None]] = [None] * len(self.symbols)
            for iat in range(len(self.symbols)):
                ifr = at2fr[iat]

                if ifr in real or ifr in ghost:
                    geom_blocks.append(self.geometry[iat])
                    symbols.append(self.symbols[iat])
                    real_atoms.append(ifr in real)
                    masses.append(self.masses[iat])

                    at2at[iat] = atom_size
                    atom_size += 1

                else:
                    at2at[iat] = None

            for ifr, fr in enumerate(self.fragments):
                if ifr in real or ifr in ghost:
                    fragments.append([at2at[iat] for iat in fr])

                if ifr in real:
                    fragment_charges.append(self.fragment_charges[ifr])
                    fragment_multiplicities.append(self.fragment_multiplicities[ifr])

                elif ifr in ghost:
                    fragment_charges.append(0)
                    fragment_multiplicities.append(1)

            assert None not in fragments

        constructor_dict["fragments"] = fragments
        constructor_dict["fragment_charges"] = fragment_charges
        constructor_dict["fragment_multiplicities"] = fragment_multiplicities
        constructor_dict["symbols"] = symbols
        constructor_dict["geometry"] = np.vstack(geom_blocks)
        constructor_dict["real"] = real_atoms
        constructor_dict["masses"] = masses

        return Molecule(orient=orient, **constructor_dict)

    def to_string(  # type: ignore
            self,
            dtype: str,
            units: str = None,
            *,
            atom_format: str = None,
            ghost_format: str = None,
            width: int = 17,
            prec: int = 12,
            return_data: bool = False):
        """Returns a string that can be used by a variety of programs.

        Unclear if this will be removed or renamed to "to_psi4_string" in the future

        Suggest psi4 --> psi4frag and psi4 route to to_string
        """
        molrec = from_schema(self.dict())
        return to_string(molrec,
                         dtype=dtype,
                         units=units,
                         atom_format=atom_format,
                         ghost_format=ghost_format,
                         width=width,
                         prec=prec,
                         return_data=return_data)

    def get_hash(self):
        """
        Returns the hash of the molecule.
        """

        m = hashlib.sha1()
        concat = ""

        tmp_dict = super().dict(skip_defaults=False)

        np.set_printoptions(precision=16)
        for field in self.hash_fields:
            data = tmp_dict[field]
            if field == "geometry":
                data = float_prep(data, GEOMETRY_NOISE)
            elif field == "fragment_charges":
                data = float_prep(data, CHARGE_NOISE)
            elif field == "molecular_charge":
                data = float_prep(data, CHARGE_NOISE)
            elif field == "masses":
                data = float_prep(data, MASS_NOISE)

            concat += json.dumps(data, default=lambda x: x.ravel().tolist())

        m.update(concat.encode("utf-8"))
        return m.hexdigest()

    def get_molecular_formula(self):
        """
        Returns the molecular formula for a molecule. Atom symbols are sorted from
        A-Z.

        Examples
        --------

        >>> methane = qcelemental.models.Molecule('''
        ... H      0.5288      0.1610      0.9359
        ... C      0.0000      0.0000      0.0000
        ... H      0.2051      0.8240     -0.6786
        ... H      0.3345     -0.9314     -0.4496
        ... H     -1.0685     -0.0537      0.1921
        ... ''')
        >>> methane.get_molecular_formula()
        CH4

        >>> hcl = qcelemental.models.Molecule('''
        ... H      0.0000      0.0000      0.0000
        ... Cl     0.0000      0.0000      1.2000
        ... ''')
        >>> hcl.get_molecular_formula()
        ClH

        """
        count = collections.Counter(x.title() for x in self.symbols)

        ret = []
        for k in sorted(count.keys()):
            c = count[k]
            ret.append(k)
            if c > 1:
                ret.append(str(c))

        return "".join(ret)

    ### Constructors

    @classmethod
    def from_data(cls,
                  data: Union[str, Dict[str, Any], np.array, bytes],
                  dtype: Optional[str] = None,
                  *,
                  orient: bool = False,
                  validate: bool = None,
                  **kwargs: Dict[str, Any]) -> 'Molecule':
        """
        Constructs a molecule object from a data structure.

        Parameters
        ----------
        data : Union[str, Dict[str, Any], np.array]
            Data to construct Molecule from
        dtype : Optional[str], optional
            How to interpret the data, if not passed attempts to discover this based on input type.
        orient : bool, optional
            Orientates the molecule to a standard frame or not.
        validate : bool, optional
            Validates the molecule or not.
        **kwargs : Dict[str, Any]
            Additional kwargs to pass to the constructors.

        Returns
        -------
        Molecule
            A constructed molecule class.

        """
        if dtype is None:
            if isinstance(data, str):
                dtype = "string"
            elif isinstance(data, np.ndarray):
                dtype = "numpy"
            elif isinstance(data, dict):
                dtype = "dict"
            elif isinstance(dtype, bytes):
                dtype = "msgpack"
            else:
                raise TypeError("Input type not understood, please supply the 'dtype' kwarg.")

        if dtype in ["string", "psi4", "psi4+", "xyz", "xyz+"]:
            mol_dict = from_string(data)
            assert isinstance(mol_dict, dict)
            input_dict = to_schema(mol_dict["qm"], dtype=2)
            validate = True
        elif dtype == "numpy":
            data = np.asarray(data)
            data = {
                "geom": data[:, 1:],
                "elez": data[:, 0],
                "units": kwargs.pop("units", "Angstrom"),
                "fragment_separators": kwargs.pop("frags", [])
            }
            input_dict = to_schema(from_arrays(**data), dtype=2)
            validate = True
        elif dtype == "msgpack":
            assert isinstance(data, bytes)
            input_dict = msgpackext_loads(data)
        elif dtype == "json":
            assert isinstance(data, str)
            input_dict = json.loads(data)
        elif dtype == "dict":
            assert isinstance(data, dict)
            input_dict = data
        else:
            raise KeyError("Dtype not understood '{}'.".format(dtype))

        return cls(orient=orient, validate=validate, **input_dict)

    @classmethod
    def from_file(cls, filename: str, dtype: Optional[str] = None, *, orient: bool = False, **kwargs):
        """
        Constructs a molecule object from a file.

        Parameters
        ----------
        filename : str
            The filename to build
        dtype : Optional[str], optional
            The type of file to interpret.
        orient : bool, optional
            Orientates the molecule to a standard frame or not.
        **kwargs
            Any additional keywords to pass to the constructor

        Returns
        -------
        Molecule
            A constructed molecule class.

        """

        ext = Path(filename).suffix

        if dtype is None:
            if ext in _extension_map:
                dtype = _extension_map[ext]
            else:
                # Let `from_string` try to sort it
                dtype = "string"

        # Raw string type, read and pass through
        if dtype in ["string", "xyz", "psi4"]:
            with open(filename, "r") as infile:
                data = infile.read()
        elif dtype == "numpy":
            data = np.load(filename)
        elif dtype == "json":
            with open(filename, "r") as infile:
                data = json.load(infile)
            dtype = "dict"
        elif dtype == "msgpack":
            with open(filename, "rb") as infile_bytes:
                data = deserialize(infile_bytes.read(), encoding="msgpack-ext")
            dtype = "dict"
        else:
            raise KeyError("Dtype not understood '{}'.".format(dtype))

        return cls.from_data(data, dtype, orient=orient, **kwargs)

    def to_file(self, filename: str, dtype: Optional[str] = None) -> None:
        """Writes the Molecule to a file.

        Parameters
        ----------
        filename : str
            The filename to write to
        dtype : Optional[str], optional
            The type of file to write, attempts to infer dtype from the filename if not provided.

        """
        ext = Path(filename).suffix

        if dtype is None:
            if ext in _extension_map:
                dtype = _extension_map[ext]
            else:
                raise KeyError(f"Could not infer dtype from filename: `{filename}`")

        flags = "w"
        if dtype in ["xyz", "psi4"]:
            stringified = self.to_string(dtype)
        elif dtype in ["json"]:
            stringified = self.serialize("json")
        elif dtype in ["msgpack", "msgpack-ext"]:
            stringified = self.serialize("msgpack-ext")
            flags = "wb"
        elif dtype in ["numpy"]:
            elements = np.array(self.atomic_numbers).reshape(-1, 1)
            npmol = np.hstack((elements, self.geometry * constants.conversion_factor("bohr", "angstroms")))
            np.save(filename, npmol)
            return

        else:
            raise KeyError(f"Dtype `{dtype}` is not valid")

        with open(filename, flags) as handle:
            handle.write(stringified)

    ### Non-Pydantic internal functions

    def _orient_molecule_internal(self):
        """
        Centers the molecule and orients via inertia tensor before returning a new set of the
        molecule geometry
        """

        new_geometry = self.geometry.copy()  # Ensure we get a copy
        # Get the mass as an array
        # Masses are needed for orientation
        np_mass = np.array(self.masses)

        # Center on Mass
        new_geometry -= np.average(new_geometry, axis=0, weights=np_mass)

        # Rotate into inertial frame
        tensor = self._inertial_tensor(new_geometry, weight=np_mass)
        _, evecs = np.linalg.eigh(tensor)

        new_geometry = np.dot(new_geometry, evecs)

        # Phases? Lets do the simplest thing and ensure the first atom in each column
        # that is not on a plane is positve

        phase_check = [False, False, False]

        geom_noise = 10**(-GEOMETRY_NOISE)
        for num in range(new_geometry.shape[0]):

            for x in range(3):
                if phase_check[x]:
                    continue

                val = new_geometry[num, x]

                if abs(val) < geom_noise:
                    continue

                phase_check[x] = True

                if val < 0:
                    new_geometry[:, x] *= -1

            if sum(phase_check) == 3:
                break
        return new_geometry

    def __str__(self) -> str:
        return self.pretty_print()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(name='{self.name}' formula='{self.get_molecular_formula()}' hash='{self.get_hash()[:7]}')>"

    def _repr_html_(self) -> str:
        try:
            return self.show()._repr_html_()
        except ModuleNotFoundError:
            return f"<p>{repr(self)[1:-1]}</p>"

    @staticmethod
    def _inertial_tensor(geom, *, weight):
        """
        Compute the moment inertia tensor for a given geometry.
        """
        # Build inertia tensor
        tensor = np.zeros((3, 3))

        # Diagonal
        tensor[0][0] = np.sum(weight * (geom[:, 1]**2.0 + geom[:, 2]**2.0))
        tensor[1][1] = np.sum(weight * (geom[:, 0]**2.0 + geom[:, 2]**2.0))
        tensor[2][2] = np.sum(weight * (geom[:, 0]**2.0 + geom[:, 1]**2.0))

        # I(alpha, beta)
        # Off diagonal
        tensor[1][0] = tensor[0][1] = -1.0 * np.sum(weight * geom[:, 0] * geom[:, 1])
        tensor[2][0] = tensor[0][2] = -1.0 * np.sum(weight * geom[:, 0] * geom[:, 2])
        tensor[2][1] = tensor[1][2] = -1.0 * np.sum(weight * geom[:, 1] * geom[:, 2])
        return tensor

    def nuclear_repulsion_energy(self, ifr: int = None) -> float:
        """Nuclear repulsion energy.

        Parameters
        ----------
        ifr : int, optional
            If not `None`, only compute for the `ifr`-th (0-indexed) fragment.

        Returns
        -------
        Nuclear repulsion energy in entire molecule or in fragment.

        """
        Zeff = [z * int(real) for z, real in zip(self.atomic_numbers, self.real)]
        atoms = list(range(self.geometry.shape[0]))

        if ifr is not None:
            atoms = self.fragments[ifr]

        nre = 0.
        for iat1, at1 in enumerate(atoms):
            for at2 in atoms[:iat1]:
                dist = np.linalg.norm(self.geometry[at1] - self.geometry[at2])
                nre += Zeff[at1] * Zeff[at2] / dist
        return nre

    def nelectrons(self, ifr: int = None) -> int:
        """Number of electrons.

        Parameters
        ----------
        ifr : int, optional
            If not `None`, only compute for the `ifr`-th (0-indexed) fragment.

        Returns
        -------
        Number of electrons in entire molecule or in fragment.

        """
        Zeff = [z * int(real) for z, real in zip(self.atomic_numbers, self.real)]

        if ifr is None:
            nel = sum(Zeff) - self.molecular_charge

        else:
            nel = sum([zf for iat, zf in enumerate(Zeff) if iat in self.fragments[ifr]]) - self.fragment_charges[ifr]

        return int(nel)

    def align(self,
              ref_mol,
              *,
              do_plot=False,
              verbose=0,
              atoms_map=False,
              run_resorting=False,
              mols_align=False,
              run_to_completion=False,
              uno_cutoff=1.e-3,
              run_mirror=False):
        """Finds shift, rotation, and atom reordering of `concern_mol` (self)
        that best aligns with `ref_mol`.

        Wraps :py:func:`qcel.molutil.B787` for :py:class:`qcel.models.Molecule`.
        Employs the Kabsch, Hungarian, and Uno algorithms to exhaustively locate
        the best alignment for non-oriented, non-ordered structures.

        Parameters
        ----------
        concern_mol : qcel.models.Molecule
            Molecule of concern, to be shifted, rotated, and reordered into
            best coincidence with `ref_mol`.
        ref_mol : qcel.models.Molecule
            Molecule to match.
        atoms_map : bool, optional
            Whether atom1 of `ref_mol` corresponds to atom1 of `concern_mol`, etc.
            If true, specifying `True` can save much time.
        mols_align : bool, optional
            Whether `ref_mol` and `concern_mol` have identical geometries by eye
            (barring orientation or atom mapping) and expected final RMSD = 0.
            If `True`, procedure is truncated when RMSD condition met, saving time.
        do_plot : bool, optional
            Pops up a mpl plot showing before, after, and ref geometries.
        run_to_completion : bool, optional
            Run reorderings to completion (past RMSD = 0) even if unnecessary because
            `mols_align=True`. Used to test worst-case timings.
        run_resorting : bool, optional
            Run the resorting machinery even if unnecessary because `atoms_map=True`.
        uno_cutoff : float, optional
            TODO
        run_mirror : bool, optional
            Run alternate geometries potentially allowing best match to `ref_mol`
            from mirror image of `concern_mol`. Only run if system confirmed to
            be nonsuperimposable upon mirror reflection.
        verbose : int, optional
            Print level.

        Returns
        -------
        Molecule, data
            Molecule is internal geometry of `self` optimally aligned and atom-ordered
              to `ref_mol`. Presently all fragment information is discarded.
            `data['rmsd']` is RMSD [A] between `ref_mol` and the optimally aligned
            geometry computed.
            `data['mill']` is a AlignmentMill with fields
            (shift, rotation, atommap, mirror) that prescribe the transformation
            from `concern_mol` and the optimally aligned geometry.

        """
        from ..molutil.align import B787

        rgeom = np.array(ref_mol.geometry)
        runiq = np.asarray([
            hashlib.sha1((sym + str(mas)).encode('utf-8')).hexdigest()
            for sym, mas in zip(ref_mol.symbols, ref_mol.masses)
        ])
        concern_mol = self
        cgeom = np.array(concern_mol.geometry)
        cmass = np.array(concern_mol.masses)
        celem = np.array(concern_mol.symbols)
        celez = np.array(concern_mol.atomic_numbers)
        cuniq = np.asarray([
            hashlib.sha1((sym + str(mas)).encode('utf-8')).hexdigest()
            for sym, mas in zip(concern_mol.symbols, concern_mol.masses)
        ])

        rmsd, solution = B787(cgeom=cgeom,
                              rgeom=rgeom,
                              cuniq=cuniq,
                              runiq=runiq,
                              do_plot=do_plot,
                              verbose=verbose,
                              atoms_map=atoms_map,
                              run_resorting=run_resorting,
                              mols_align=mols_align,
                              run_to_completion=run_to_completion,
                              run_mirror=run_mirror,
                              uno_cutoff=uno_cutoff)

        ageom, amass, aelem, aelez, _ = solution.align_system(cgeom, cmass, celem, celez, cuniq, reverse=False)
        adict = from_arrays(geom=ageom,
                            mass=amass,
                            elem=aelem,
                            elez=aelez,
                            units='Bohr',
                            molecular_charge=concern_mol.molecular_charge,
                            molecular_multiplicity=concern_mol.molecular_multiplicity,
                            fix_com=True,
                            fix_orientation=True)
        amol = Molecule(validate=False, **to_schema(adict, dtype=2))

        # TODO -- can probably do more with fragments in amol now that
        #         Mol is something with non-contig frags. frags now discarded.

        assert compare_values(concern_mol.nuclear_repulsion_energy(),
                              amol.nuclear_repulsion_energy(),
                              'Q: concern_mol-->returned_mol NRE uncorrupted',
                              atol=1.e-4,
                              quiet=(verbose > 1))
        if mols_align:
            assert compare_values(ref_mol.nuclear_repulsion_energy(),
                                  amol.nuclear_repulsion_energy(),
                                  'Q: concern_mol-->returned_mol NRE matches ref_mol',
                                  atol=1.e-4,
                                  quiet=(verbose > 1))
            assert compare(True,
                           np.allclose(ref_mol.geometry, amol.geometry, atol=4),
                           'Q: concern_mol-->returned_mol geometry matches ref_mol',
                           quiet=(verbose > 1))

        return amol, {'rmsd': rmsd, 'mill': solution}

    def scramble(self,
                 *,
                 do_shift: bool = True,
                 do_rotate=True,
                 do_resort=True,
                 deflection=1.0,
                 do_mirror=False,
                 do_plot=False,
                 do_test=False,
                 run_to_completion=False,
                 run_resorting=False,
                 verbose=0):
        """Generate a Molecule with random or directed translation, rotation, and atom shuffling.
        Optionally, check that the aligner returns the opposite transformation.

        Parameters
        ----------
        ref_mol : qcel.models.Molecule
            Molecule to perturb.
        do_shift : bool or array-like, optional
            Whether to generate a random atom shift on interval [-3, 3) in each
            dimension (`True`) or leave at current origin. To shift by a specified
            vector, supply a 3-element list.
        do_rotate : bool or array-like, optional
            Whether to generate a random 3D rotation according to algorithm of Arvo.
            To rotate by a specified matrix, supply a 9-element list of lists.
        do_resort : bool or array-like, optional
            Whether to shuffle atoms (`True`) or leave 1st atom 1st, etc. (`False`).
            To specify shuffle, supply a nat-element list of indices.
        deflection : float, optional
            If `do_rotate`, how random a rotation: 0.0 is no change, 0.1 is small
            perturbation, 1.0 is completely random.
        do_mirror : bool, optional
            Whether to construct the mirror image structure by inverting y-axis.
        do_plot : bool, optional
            Pops up a mpl plot showing before, after, and ref geometries.
        do_test : bool, optional
            Additionally, run the aligner on the returned Molecule and check that
            opposite transformations obtained.
        run_to_completion : bool, optional
            By construction, scrambled systems are fully alignable (final RMSD=0).
            Even so, `True` turns off the mechanism to stop when RMSD reaches zero
            and instead proceed to worst possible time.
        run_resorting : bool, optional
            Even if atoms not shuffled, test the resorting machinery.
        verbose : int, optional
            Print level.

        Returns
        -------
        Molecule, data
            Molecule is scrambled copy of `ref_mol` (self).
            `data['rmsd']` is RMSD [A] between `ref_mol` and the scrambled geometry.
            `data['mill']` is a AlignmentMill with fields
            (shift, rotation, atommap, mirror) that prescribe the transformation
            from `ref_mol` to the returned geometry.

        Raises
        ------
        AssertionError
            If `do_test=True` and aligner sanity check fails for any of the reverse
            transformations.

        """
        from ..molutil.align import compute_scramble

        ref_mol = self
        rgeom = np.array(ref_mol.geometry)
        rmass = np.array(ref_mol.masses)
        relem = np.array(ref_mol.symbols)
        relez = np.array(ref_mol.atomic_numbers)
        runiq = np.asarray([
            hashlib.sha1((sym + str(mas)).encode('utf-8')).hexdigest()
            for sym, mas in zip(ref_mol.symbols, ref_mol.masses)
        ])
        nat = rgeom.shape[0]

        perturbation = compute_scramble(rgeom.shape[0],
                                        do_shift=do_shift,
                                        do_rotate=do_rotate,
                                        deflection=deflection,
                                        do_resort=do_resort,
                                        do_mirror=do_mirror)
        cgeom, cmass, celem, celez, cuniq = perturbation.align_system(rgeom, rmass, relem, relez, runiq, reverse=True)
        cmolrec = from_arrays(
            geom=cgeom,
            mass=cmass,
            elem=celem,
            elez=celez,
            units='Bohr',
            molecular_charge=ref_mol.molecular_charge,
            molecular_multiplicity=ref_mol.molecular_multiplicity,
            # copying fix_* vals rather than outright True. neither way great
            fix_com=ref_mol.fix_com,
            fix_orientation=ref_mol.fix_orientation)
        cmol = Molecule(validate=False, **to_schema(cmolrec, dtype=2))

        rmsd = np.linalg.norm(cgeom - rgeom) * constants.bohr2angstroms / np.sqrt(nat)
        if verbose >= 1:
            print('Start RMSD = {:8.4f} [A]'.format(rmsd))

        if do_test:
            _, data = cmol.align(ref_mol,
                                 do_plot=do_plot,
                                 atoms_map=(not do_resort),
                                 run_resorting=run_resorting,
                                 mols_align=True,
                                 run_to_completion=run_to_completion,
                                 run_mirror=do_mirror,
                                 verbose=verbose)
            solution = data['mill']

            assert compare(True,
                           np.allclose(solution.shift, perturbation.shift, atol=6),
                           'shifts equiv',
                           quiet=(verbose > 1))
            if not do_resort:
                assert compare(True,
                               np.allclose(solution.rotation.T, perturbation.rotation),
                               'rotations transpose',
                               quiet=(verbose > 1))
            if solution.mirror:
                assert compare(True, do_mirror, 'mirror allowed', quiet=(verbose > 1))

        return cmol, {'rmsd': rmsd, 'mill': perturbation}


# auto_gen_docs_on_demand(Molecule)
